import os
import shutil
import time
import warnings
from abc import abstractmethod
from pathlib import Path
from typing import Any, Optional, Sequence

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt

from CADETProcess import CADETProcessError, log, settings
from CADETProcess.dataStructure import (
    RangedInteger,
    Structure,
    Typed,
    UnsignedFloat,
    UnsignedInteger,
)
from CADETProcess.optimization import (
    Joblib,
    OptimizationProblem,
    OptimizationResults,
    ParallelizationBackendBase,
    Population,
)

__all__ = ["OptimizerBase"]


class OptimizerBase(Structure):
    """
    BaseClass for optimization solver APIs.

    Holds the configuration of the individual solvers and gives an interface
    for calling the run method. The class has to convert the
    OptimizationProblem configuration to the APIs configuration format and
    convert the results back to the CADET-Process format.

    Attributes
    ----------
    is_population_based : bool
        True, if the optimizer evaluates entire populations at every step.
    supports_multi_objective : bool
        True, if the optimizer supports multi-objective optimization.
    supports_linear_constraints : bool
        True, if the optimizer supports linear constraints.
    supports_linear_equality_constraints : bool
        True, if the optimizer supports linear equality constraints.
    supports_nonlinear_constraints : bool
        True, if the optimizer supports nonlinear constraints.
    supports_bounds : bool
        True, if the optimizer supports bound constraints
    ignore_linear_constraints_config: bool
        True, if the optimizer can handle transforms and dependent variables in linear
        constraints.
    progress_frequency : int
        Number of generations after which the optimizer reports progress.
    cv_bounds_tol : float
        Tolerance for bounds constraint violation.
        The default is 0.0.
    cv_lincon_tol : float
        Tolerance for linear constraints violation.
        The default is 0.0.
    cv_lineqcon_tol : float
        Tolerance for linear equality constraints violation.
        The default is 0.0.
    cv_nonlincon_tol : float
        Tolerance for nonlinear constraints violation.
        The default is 0.0.
    similarity_tol : UnsignedFloat, optional
        Tolerance for individuals to be considered similar.
        Similar items are removed from the Pareto front to limit its size.
        The default is None, indicating that all individuals should be kept.
    n_max_evals : int, optional
        Maximum number of function evaluations.
    n_max_iter : int, optional
        Maximum number of iterations (e.g. generations).
    parallelization_backend : ParallelizationBackendBase, optional
        Class used to handle parallelized (and also sequential) evaluation of eval_fun
        functions for each individual in a given population.
        The default parallelization backend is 'Joblib', which provides parallel
        execution using multiple cores.
    n_cores : int, optional
        Proxy to the number of cores used by the parallelization backend.
    """

    is_population_based = False

    supports_single_objective = True
    supports_multi_objective = False
    supports_linear_constraints = False
    supports_linear_equality_constraints = False
    supports_nonlinear_constraints = False
    supports_bounds = False

    ignore_linear_constraints_config = False

    progress_frequency = RangedInteger(lb=1)

    x_tol = UnsignedFloat()
    f_tol = UnsignedFloat()

    cv_bounds_tol = UnsignedFloat(default=0.0)
    cv_lineqcon_tol = UnsignedFloat(default=0.0)
    cv_lincon_tol = UnsignedFloat(default=0.0)
    cv_nonlincon_tol = UnsignedFloat(default=0.0)

    n_max_iter = UnsignedInteger(default=100000)
    n_max_evals = UnsignedInteger(default=100000)

    similarity_tol = UnsignedFloat()
    parallelization_backend = Typed(ty=ParallelizationBackendBase)

    _general_options = [
        "progress_frequency",
        "x_tol",
        "f_tol",
        "cv_bounds_tol",
        "cv_lincon_tol",
        "cv_lineqcon_tol",
        "cv_nonlincon_tol",
        "n_max_iter",
        "n_max_evals",
        "similarity_tol",
    ]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize OptimizerBase."""
        self.parallelization_backend = Joblib()

        super().__init__(*args, **kwargs)

    def optimize(
        self,
        optimization_problem: OptimizationProblem,
        x0: Optional[list] = None,
        save_results: Optional[bool] = True,
        results_directory: Optional[str] = None,
        use_checkpoint: Optional[bool] = False,
        overwrite_results_directory: Optional[bool] = False,
        exist_ok: Optional[bool] = True,
        log_level: Optional[str] = "INFO",
        reinit_cache: Optional[bool] = True,
        delete_cache: bool = True,
        *args: Any,
        **kwargs: Any,
    ) -> OptimizationResults:
        """
        Solve OptimizationProblem.

        Parameters
        ----------
        optimization_problem : OptimizationProblem
            OptimizationProblem to be solved.
        x0 : list, optional
            Initial values. If None, valid points are generated.
        save_results : bool, optional
            If True, save results. The default is True.
        results_directory : str, optional
            Results directory. If None, working directory is used.
            Only has an effect, if save_results == True.
        use_checkpoint : bool, optional
            If True, try continuing fom checkpoint. The default is True.
            Only has an effect, if save_results == True.
        overwrite_results_directory : bool, optional
            If True, overwrite existing results directory. The default is False.
        exist_ok : bool, optional
            If False, Exception is raised when results_directory is not empty.
            The default is True.
        log_level : str, optional
            log level. The default is "INFO".
        reinit_cache : bool, optional
            If True, reinitialize the Cache. The default is True.
        delete_cache : bool, optional
            If True, delete ResultsCache after finishing. The default is True.
        *args : TYPE
            Additional arguments for Optimizer.
        **kwargs : TYPE
            Additional keyword arguments for Optimizer.

        Raises
        ------
        TypeError
            If optimization_problem is not an instance of OptimizationProblem.
        CADETProcessError
            If Optimizer is not suited for OptimizationProblem (e.g. multi-objective).

        Returns
        -------
        results : OptimizationResults
            Results of the Optimization.

        See Also
        --------
        OptimizationProblem
        OptimizationResults
        CADETProcess.optimization.ResultsCache
        """
        self._current_cache_entries = []

        self.logger = log.get_logger(str(self), level=log_level)

        # Check OptimizationProblem
        if not isinstance(optimization_problem, OptimizationProblem):
            raise TypeError("Expected OptimizationProblem")

        if not self.check_optimization_problem(optimization_problem):
            raise CADETProcessError("Cannot solve OptimizationProblem.")

        self.optimization_problem = optimization_problem

        # Setup OptimizationResults
        self.results = OptimizationResults(
            optimization_problem=optimization_problem,
            optimizer=self,
            similarity_tol=self.similarity_tol,
        )

        if save_results:
            if results_directory is None:
                results_directory = (
                    settings.working_directory / f"results_{optimization_problem.name}"
                )
            results_directory = Path(results_directory)
            if overwrite_results_directory and results_directory.exists():
                shutil.rmtree(results_directory)
            try:
                results_directory.mkdir(exist_ok=exist_ok, parents=True)
            except FileExistsError:
                raise CADETProcessError(
                    "Results directory already exists. "
                    "To continue using same directory, 'exist_ok=True'. "
                    "To overwrite, set 'overwrite_results_directory=True. "
                )

            self.results.results_directory = results_directory

            checkpoint_path = os.path.join(results_directory, "checkpoint.h5")
            if use_checkpoint and os.path.isfile(checkpoint_path):
                self.logger.info("Continue optimization from checkpoint.")
                self.results.load_results(checkpoint_path)
            else:
                self.results.setup_csv()

        # Setup Callbacks
        if save_results and optimization_problem.n_callbacks > 0:
            callbacks_dir = results_directory / "callbacks"
            callbacks_dir.mkdir(exist_ok=True)

            if optimization_problem.n_callbacks > 1:
                for callback in optimization_problem.callbacks:
                    callback_dir = callbacks_dir / str(callback)
                    callback_dir.mkdir(exist_ok=True)
        else:
            callbacks_dir = None
        self.callbacks_dir = callbacks_dir

        if reinit_cache:
            self.optimization_problem.setup_cache(self.n_cores)

        if x0 is not None:
            flag, x0 = self.check_x0(optimization_problem, x0)

            if not flag:
                raise ValueError("x0 contains invalid entries.")

        log.log_time("Optimization", self.logger.level)(self._run)
        log.log_results("Optimization", self.logger.level)(self._run)
        log.log_exceptions("Optimization", self.logger.level)(self._run)

        backend = plt.get_backend()
        plt.switch_backend("agg")

        start = time.time()
        self._run(self.optimization_problem, x0, *args, **kwargs)
        time_elapsed = time.time() - start

        self.results.time_elapsed = time_elapsed
        self.results.cpu_time = self.n_cores * time_elapsed

        self.run_final_processing()

        if delete_cache:
            optimization_problem.delete_cache(reinit=True)
        self._current_cache_entries = []

        plt.switch_backend(backend)

        if not self.results.success:
            raise CADETProcessError(
                f"Optimizaton failed with message: {self.results.exit_message}"
            )

        return self.results

    def load_results(
        self,
        checkpoint_path: str | Path,
        optimization_problem: Optional[OptimizationProblem] = None,
    ) -> OptimizationResults:
        """
        Load optimization results from a checkpoint file.

        Parameters
        ----------
        checkpoint_path : str | Path
            Path to the checkpoint file.
        optimization_problem : OptimizationProblem, optional
            The optimization problem associated with the results.
            If None, results are loaded without a problem reference.

        Returns
        -------
        OptimizationResults
            The loaded optimization results.
        """
        results = OptimizationResults(
            optimization_problem=optimization_problem,
            optimizer=self,
            similarity_tol=self.similarity_tol,
        )

        results.load_results(checkpoint_path)

        return results

    @abstractmethod
    def _run(
        optimization_problem, x0: Optional[list] = None, *args: Any, **kwargs: Any
    ) -> None:
        """
        Abstract Method for solving an optimization problem.

        Parameters
        ----------
        optimization_problem : OptimizationProblem
            Optimization problem to be solved.
        x0 : list, optional
            Initial population of independent variables in untransformed space.
        *args : args, optional
            Additional args that may be processed.
        **kwargs : kwargs, optional
            Additional kwargs that may be processed.

        Returns
        -------
        results : OptimizationResults
            Optimization results including OptimizationProblem and Optimizer
            configuration.

        Raises
        ------
        CADETProcessError
            If solver doesn't terminate successfully
        """
        return

    def check_optimization_problem(
        self, optimization_problem: OptimizationProblem
    ) -> bool:
        """
        Check if problem is configured correctly and supported by the optimizer.

        Parameters
        ----------
        optimization_problem: OptimizationProblem
            An optimization problem to check.

        Returns
        -------
        flag : bool
            True if the optimization problem is supported and configured correctly,
            False otherwise.
        """
        flag = True
        if not optimization_problem.check_config(
            ignore_linear_constraints=self.ignore_linear_constraints_config
        ):
            # Warnings are raised internally
            flag = False

        if (
            optimization_problem.n_objectives == 1
            and not self.supports_single_objective
        ):
            warnings.warn("Optimizer does not support single-objective problems")
            flag = False

        if optimization_problem.n_objectives > 1 and not self.supports_multi_objective:
            warnings.warn("Optimizer does not support multi-objective problems")
            flag = False

        if (
            not np.all(
                np.isinf(optimization_problem.lower_bounds_independent_transformed)
            )
            and not np.all(
                np.isinf(optimization_problem.upper_bounds_independent_transformed)
            )
        ) and not self.supports_bounds:
            warnings.warn("Optimizer does not support bounds")
            flag = False

        if (
            optimization_problem.n_linear_constraints > 0
            and not self.supports_linear_constraints
        ):
            warnings.warn(
                "Optimizer does not support problems with linear constraints."
            )
            flag = False

        if (
            optimization_problem.n_linear_equality_constraints > 0
            and not self.supports_linear_equality_constraints
        ):
            warnings.warn(
                "Optimizer does not support problems with linear equality constraints."
            )
            flag = False

        if (
            optimization_problem.n_nonlinear_constraints > 0
            and not self.supports_nonlinear_constraints
        ):
            warnings.warn(
                "Optimizer does not support problems with nonlinear constraints."
            )
            flag = False

        return flag

    def check_x0(
        self, optimization_problem: OptimizationProblem, x0: npt.ArrayLike
    ) -> tuple:
        """
        Check the initial guess x0 for an optimization problem.

        Parameters
        ----------
        optimization_problem : OptimizationProblem
            The optimization problem instance to which x0 is related.
        x0 : array_like
            The initial guess for the optimization problem.
            It can be a single individual or a population.

        Returns
        -------
        tuple
            A tuple containing a boolean flag indicating if x0 is valid, and the
            potentially modified x0.
        """
        flag = True

        shape = np.array(x0).shape

        is_x0_1d = len(shape) == 1

        x0 = np.array(x0, ndmin=2)

        n_dependent_variables = optimization_problem.n_dependent_variables
        n_independent_variables = optimization_problem.n_independent_variables
        n_variables = n_dependent_variables + n_independent_variables

        if x0.shape[1] != n_variables and x0.shape[1] != n_independent_variables:
            warnings.warn(
                f"x0 for optimization problem is expected to be of length "
                f"{n_independent_variables} or"
                f"{n_variables}. Got {x0.shape[1]}"
            )
            flag = False

        if n_dependent_variables > 0 and x0.shape[1] == n_variables:
            x0 = [optimization_problem.get_independent_values(ind) for ind in x0]
            warnings.warn(
                "x0 contains dependent values. "
                "Will recompute dependencies for consistency."
            )
            x0 = np.array(x0)

        for x in x0:
            if not optimization_problem.check_individual(
                x,
                get_dependent_values=True,
                cv_bounds_tol=self.cv_bounds_tol,
                cv_lincon_tol=self.cv_lincon_tol,
                cv_lineqcon_tol=self.cv_lineqcon_tol,
                check_nonlinear_constraints=False,
                silent=True,
            ):
                flag = False
                break

        if is_x0_1d:
            x0 = x0[0]

        x0 = x0.tolist()

        return flag, x0

    def _create_population(
        self,
        X_transformed: npt.ArrayLike,
        F: npt.ArrayLike,
        F_min: npt.ArrayLike,
        G: npt.ArrayLike,
        CV_nonlincon: npt.ArrayLike,
    ) -> Population:
        """Create new population from current generation for post procesing."""
        X_transformed = np.array(X_transformed, ndmin=2)
        F = np.array(F, ndmin=2)
        F_min = np.array(F_min, ndmin=2)
        G = np.array(G, ndmin=2)
        CV_nonlincon = np.array(CV_nonlincon, ndmin=2)

        if self.optimization_problem.n_meta_scores > 0:
            M_min = self.optimization_problem.evaluate_meta_scores(
                X_transformed,
                untransform=True,
                ensure_minimization=True,
                parallelization_backend=self.parallelization_backend,
            )
            M = self.optimization_problem.transform_maximization(
                M_min, scores="meta_scores"
            )
        else:
            M_min = None
            M = None

        if self.optimization_problem.n_nonlinear_constraints == 0:
            G = None
            CV_nonlincon = None

        X = self.optimization_problem.get_dependent_values(
            X_transformed, untransform=True
        )
        population = self.optimization_problem.create_population(
            X,
            F=F,
            F_min=F_min,
            G=G,
            CV_nonlincon=CV_nonlincon,
            M=M,
            M_min=M_min,
        )

        for ind in population:
            ind.is_feasible = self.optimization_problem.check_individual(
                ind.x,
                cv_bounds_tol=self.cv_bounds_tol,
                cv_lincon_tol=self.cv_lincon_tol,
                cv_lineqcon_tol=self.cv_lineqcon_tol,
                check_nonlinear_constraints=True,
                cv_nonlincon_tol=self.cv_nonlincon_tol,
                silent=True,
            )

        return population

    def _create_pareto_front(self, X_opt_transformed: npt.ArrayLike) -> Population:
        """Create new pareto front from current generation for post procesing."""
        if X_opt_transformed is None:
            pareto_front = None
        else:
            pareto_front = Population()

            for x_opt_transformed in X_opt_transformed:
                x_opt = self.optimization_problem.get_dependent_values(
                    x_opt_transformed, untransform=True
                )
                ind = self.results.population_all[x_opt]
                pareto_front.add_individual(ind)

        return pareto_front

    def _create_meta_front(self) -> Population:
        """Create new meta front from current generation for post procesing."""
        if self.optimization_problem.n_multi_criteria_decision_functions == 0:
            meta_front = None
        else:
            pareto_front = self.results.pareto_front

            X_meta_front = (
                self.optimization_problem.evaluate_multi_criteria_decision_functions(
                    pareto_front
                )
            )

            meta_front = Population()
            for x in X_meta_front:
                meta_front.add_individual(pareto_front[x])

        return meta_front

    def _evaluate_callbacks(
        self,
        current_generation: int,
        sub_dir: str = None,
    ) -> None:
        if sub_dir is not None:
            callbacks_dir = self.callbacks_dir / sub_dir
            callbacks_dir.mkdir(exist_ok=True, parents=True)
        else:
            callbacks_dir = self.callbacks_dir

        for callback in self.optimization_problem.callbacks:
            if self.optimization_problem.n_callbacks > 1:
                _callbacks_dir = callbacks_dir / str(callback)
                _callbacks_dir.mkdir(exist_ok=True, parents=True)
            else:
                _callbacks_dir = callbacks_dir

            callback.cleanup(_callbacks_dir, current_generation)
            callback._callbacks_dir = _callbacks_dir

        self.optimization_problem.evaluate_callbacks(
            self.results.meta_front,
            current_generation,
            parallelization_backend=self.parallelization_backend,
        )

    def _log_results(self, current_generation: int) -> None:
        self.logger.info(f"Finished Generation {current_generation}.")
        for ind in self.results.meta_front:
            message = f"x: {ind.x}, f: {ind.f}"

            if self.optimization_problem.n_nonlinear_constraints > 0:
                message += f", cv: {ind.cv_nonlincon}"

            if self.optimization_problem.n_meta_scores > 0:
                message += f", m: {ind.m}"
            self.logger.info(message)

    def run_post_processing(
        self,
        X_transformed: Sequence[Sequence[float]],
        F_minimized: Sequence[float | Sequence[float]],
        G: Sequence[float | Sequence[float]],
        CV_nonlincon: Sequence[float],
        current_generation: int,
        X_opt_transformed: Optional[Sequence[float]] = None,
    ) -> None:
        """
        Run post-processing of generation.

        Notes
        -----
        This method also works for optimizers that only perform a single evaluation per
        "generation".

        Parameters
        ----------
        X_transformed : list
            Optimization variable values of generation in independent transformed space.
        F_minimized : list
            Objective function values of generation.
            This assumes that all objective function values are minimized.
        G : list
            Nonlinear constraint function values of generation.
        CV_nonlincon : list
            Nonlinear constraints violation of of generation.
        current_generation : int
            Current generation.
        X_opt_transformed : list, optional
            (Currently) best variable values in independent transformed space.
            If None, internal pareto front is used to determine best values.
        """
        F = self.optimization_problem.transform_maximization(
            F_minimized, scores="objectives"
        )
        population = self._create_population(
            X_transformed, F, F_minimized, G, CV_nonlincon
        )
        self.results.update(population)

        pareto_front = self._create_pareto_front(X_opt_transformed)
        self.results.update_pareto(pareto_front)

        meta_front = self._create_meta_front()
        if meta_front is not None:
            self.results.update_meta(meta_front)

        if (
            self.progress_frequency is not None
            and current_generation % self.progress_frequency == 0
        ):
            self.results.plot_figures(show=False)

        self._evaluate_callbacks(current_generation)

        self.results.save_results("checkpoint")

        # Remove new entries from cache that didn't make it to the meta front
        for x in population.x:
            x_key = x.tobytes()
            if x not in self.results.meta_front.x:
                self.optimization_problem.prune_cache(x_key, close=False)
            else:
                self._current_cache_entries.append(x_key)

        # Remove old meta front entries from cache that were replaced by better ones
        for x_key in self._current_cache_entries:
            x = np.frombuffer(x_key)
            if not np.all(np.isin(x, self.results.meta_front.x)):
                self.optimization_problem.prune_cache(x_key, close=False)
                self._current_cache_entries.remove(x_key)

        self._log_results(current_generation)

    def run_final_processing(self) -> None:
        """Run post processing at the end of the optimization."""
        self.results.plot_figures(show=False)
        if self.optimization_problem.n_callbacks > 0:
            self._evaluate_callbacks(0, "final")
        self.results.save_results("final")

    @property
    def options(self) -> dict:
        """dict: Optimizer options."""
        return {
            opt: getattr(self, opt)
            for opt in (self._general_options + self._specific_options)
        }

    @property
    def specific_options(self) -> dict:
        """dict: Optimizer spcific options."""
        return {opt: getattr(self, opt) for opt in (self._specific_options)}

    @property
    def n_cores(self) -> int:
        """int: Proxy to the number of cores used by the parallelization backend.

        Note, this will always return the actual number of cores used, even if negative
        values are set.

        See Also
        --------
        parallelization_backend
        """
        return self.parallelization_backend._n_cores

    @n_cores.setter
    def n_cores(self, n_cores: int) -> None:
        self.parallelization_backend.n_cores = n_cores

    def __str__(self) -> str:
        """str: String representation."""
        return self.__class__.__name__



from collections import defaultdict
from dataclasses import dataclass
from typing import Any, Literal, Optional
from warnings import warn

import numpy as np
from addict import Dict
from scipy import integrate, interpolate

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import cached_property_if_locked
from CADETProcess.dynamicEvents import EventHandler, Section, TimeLine

from .componentSystem import ComponentSystem
from .flowSheet import FlowSheet
from .unitOperation import Inlet, Outlet


class Process(EventHandler):
    """
    Class for defining the dynamic changes of a flow sheet.

    Attributes
    ----------
    name : str
        Name of the process object to be simulated.
    system_state : np.ndarray
        State of the process object
    system_state_derivate : np.ndarray
        Derivative of the state

    See Also
    --------
    EventHandler
    CADETProcess.processModel.FlowSheet
    CADETProcess.simulation.Solver
    """

    _initial_states = ["system_state", "system_state_derivative"]

    def __init__(
        self,
        flow_sheet: FlowSheet,
        name: str,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize Process."""
        self.flow_sheet = flow_sheet
        self.name = name

        self.system_state = None
        self.system_state_derivative = None

        self._parameter_sensitivities = []

        self._meta_information = Dict()

        super().__init__(*args, **kwargs)

    @property
    def n_comp(self) -> int:
        """int: Number of components in the process."""
        return self.flow_sheet.n_comp

    @property
    def meta_information(self) -> dict:
        """dict: Meta information of the process."""
        # TODO: DO we still use this anywhere?
        return self._meta_information

    @property
    def component_system(self) -> ComponentSystem:
        """ComponentSystem: Component system of the process."""
        return self.flow_sheet.component_system

    @property
    def flow_sheet(self) -> FlowSheet:
        """FlowSheet: flow sheet of the process model.

        Raises
        ------
        TypeError:
            If flow_sheet is not an instance of FlowSheet.

        """
        return self._flow_sheet

    @flow_sheet.setter
    def flow_sheet(self, flow_sheet: FlowSheet) -> None:
        if not isinstance(flow_sheet, FlowSheet):
            raise TypeError("Expected FlowSheet")
        self._flow_sheet = flow_sheet

    @cached_property_if_locked
    def m_feed(self) -> np.ndarray:
        """np.ndarray: Mass of feed components entering the system in one cycle."""
        flow_rate_timelines = self.flow_rate_timelines

        feed_all = np.zeros((self.n_comp,))
        for feed in self.flow_sheet.feed_inlets:
            feed_flow_rate_time_line = flow_rate_timelines[feed.name].total_out[None]
            feed_signal_param = f"flow_sheet.{feed.name}.c"
            if feed_signal_param in self.parameter_timelines:
                tl = self.parameter_timelines[feed_signal_param]
                feed_signal_time_line = tl
            else:
                feed_signal_time_line = TimeLine()
                feed_section = Section(0, self.cycle_time, feed.c, is_polynomial=True)
                feed_signal_time_line.add_section(feed_section)

            m_i = [
                integrate.quad(
                    lambda t: feed_flow_rate_time_line.value(t) * feed_signal_time_line.value(t)[comp],  # noqa: E501
                    0,
                    self.cycle_time,
                    points=self.event_times,
                )[0]
                for comp in range(self.n_comp)
            ]

            feed_all += np.array(m_i)

        return feed_all

    @cached_property_if_locked
    def V_eluent(self) -> float:
        """float: Volume of the eluent entering the system in one cycle."""
        flow_rate_timelines = self.flow_rate_timelines

        V_all = 0
        for eluent in self.flow_sheet.eluent_inlets:
            eluent_time_line = flow_rate_timelines[eluent.name]["total_out"][None]
            V_eluent = eluent_time_line.integral().squeeze()
            V_all += V_eluent

        return float(V_all)

    @cached_property_if_locked
    def V_solid(self) -> float:
        """float: Volume of all solid phase material used in flow sheet."""
        return sum([unit.volume_solid for unit in self.flow_sheet.units_with_binding])

    @cached_property_if_locked
    def flow_rate_timelines(self) -> dict:
        """Return TimeLine of flow_rate for all unit_operations."""
        flow_rate_timelines = {
            unit.name: {
                "total_in": defaultdict(TimeLine),
                "origins": defaultdict(
                    lambda: defaultdict(lambda: defaultdict(TimeLine))
                ),
                "total_out": defaultdict(TimeLine),
                "destinations": defaultdict(
                    lambda: defaultdict(lambda: defaultdict(TimeLine))
                ),
            }
            for unit in self.flow_sheet.units
        }

        # Create dummy section state for Processes without events
        if len(self.section_states) == 0:
            it = [(None, {})]
        else:
            it = self.section_states.items()

        for i, (time, state) in enumerate(it):
            start = self.section_times[i]
            end = self.section_times[i + 1]

            flow_rates = self.flow_sheet.get_flow_rates(state)

            for unit, flow_rate_dict in flow_rates.items():
                unit_flow_rates = flow_rate_timelines[unit]

                # If inlet, also use outlet for total_in

                if isinstance(self.flow_sheet[unit], Inlet):
                    for port in flow_rate_dict["total_out"]:
                        section = Section(
                            start,
                            end,
                            flow_rate_dict.total_out[port],
                            is_polynomial=True,
                        )
                        unit_flow_rates["total_in"][port].add_section(section)
                else:
                    for port in flow_rate_dict["total_in"]:
                        section = Section(
                            start,
                            end,
                            flow_rate_dict.total_in[port],
                            is_polynomial=True,
                        )
                        unit_flow_rates["total_in"][port].add_section(section)

                for port in flow_rate_dict.origins:
                    for orig, origin_port_dict in flow_rate_dict.origins[port].items():
                        for orig_port, flow_rate_orig in origin_port_dict.items():
                            section = Section(
                                start, end, flow_rate_orig, is_polynomial=True
                            )
                            unit_flow_rates["origins"][port][orig][orig_port].add_section(section)

                # If outlet, also use inlet for total_out
                if isinstance(self.flow_sheet[unit], Outlet):
                    for port in flow_rate_dict["total_in"]:
                        section = Section(
                            start,
                            end,
                            flow_rate_dict.total_in[port],
                            is_polynomial=True,
                        )
                        unit_flow_rates["total_out"][port].add_section(section)
                else:
                    for port in flow_rate_dict["total_out"]:
                        section = Section(
                            start,
                            end,
                            flow_rate_dict.total_out[port],
                            is_polynomial=True,
                        )
                        unit_flow_rates["total_out"][port].add_section(section)

                for port in flow_rate_dict.destinations:
                    for dest, dest_port_dict in flow_rate_dict.destinations[port].items():
                        for dest_port, flow_rate_dest in dest_port_dict.items():
                            section = Section(
                                start,
                                end,
                                flow_rate_dest,
                                is_polynomial=True,
                            )
                            unit_flow_rates["destinations"][port][dest][dest_port].add_section(section)

        return Dict(flow_rate_timelines)

    @cached_property_if_locked
    def flow_rate_section_states(self) -> dict:
        """Return flow rates for all units for every section time."""
        section_states = {
            time: {
                unit.name: {
                    "total_in": defaultdict(list),
                    "origins": defaultdict(
                        lambda: defaultdict(lambda: defaultdict(list))
                    ),
                    "total_out": defaultdict(list),
                    "destinations": defaultdict(
                        lambda: defaultdict(lambda: defaultdict(list))
                    ),
                }
                for unit in self.flow_sheet.units
            }
            for time in self.section_times[0:-1]
        }

        for sec_time in self.section_times[0:-1]:
            for unit, unit_flow_rates in self.flow_rate_timelines.items():
                if isinstance(self.flow_sheet[unit], Inlet):
                    for port in unit_flow_rates["total_out"]:
                        section_states[sec_time][unit]["total_in"][port] = \
                        unit_flow_rates["total_out"][port].coefficients(sec_time)
                else:
                    for port in unit_flow_rates["total_in"]:
                        section_states[sec_time][unit]["total_in"][port] = \
                            unit_flow_rates["total_in"][port].coefficients(sec_time)

                    for port, orig_dict in unit_flow_rates.origins.items():
                        for origin in orig_dict:
                            for origin_port, tl in orig_dict[origin].items():
                                section_states[sec_time][unit]["origins"][port][origin][origin_port] = \
                                    tl.coefficients(sec_time)  # noqa: E501

                if isinstance(self.flow_sheet[unit], Outlet):
                    for port in unit_flow_rates["total_in"]:
                        section_states[sec_time][unit]["total_out"][port] = \
                            unit_flow_rates["total_in"][port].coefficients(sec_time)
                else:
                    for port in unit_flow_rates["total_out"]:
                        section_states[sec_time][unit]["total_out"][port] = \
                            unit_flow_rates["total_out"][port].coefficients(sec_time)

                    for port, dest_dict in unit_flow_rates.destinations.items():
                        for dest in dest_dict:
                            for dest_port, tl in dest_dict[dest].items():
                                section_states[sec_time][unit]["destinations"][port][dest][dest_port] = \
                                    tl.coefficients(sec_time)  # noqa: E501

        return Dict(section_states)

    @property
    def n_sensitivities(self) -> int:
        """int: Number of parameter sensitivities."""
        return len(self.parameter_sensitivities)

    @property
    def parameter_sensitivities(self) -> list:
        """list: Parameter sensitivites."""
        return self._parameter_sensitivities

    @property
    def parameter_sensitivity_names(self) -> list:
        """list: Parameter sensitivity names."""
        return [sens.name for sens in self.parameter_sensitivities]

    def add_parameter_sensitivity(
        self,
        parameter_paths: str | list[str],
        name: Optional[str] = None,
        components: Optional[str | list[str]] = None,
        polynomial_coefficients: Optional[str | list[str]] = None,
        reaction_indices: Optional[int | list[int]] = None,
        bound_state_indices: Optional[int | list[int]] = None,
        section_indices: Optional[int | list[int]] = None,
        abstols: Optional[float | list[float]] = None,
        factors: Optional[int | list[int]] = None,
    ) -> None:
        """
        Add parameter sensitivty to Process.

        Parameters
        ----------
        parameter_paths : str or list of str
            The path to the parameter(s).
        name : str, optional
            The name of the parameter sensitivity.
            If not provided, the name of the first parameter will be used.
        components : str or list of str, optional
            The component(s) to which the parameter(s) belong.
            Must only be provided if parameter is specific to a certain compoment.
        polynomial_coefficients: str or list of str, optional
            The polynomial coefficients(s) to which the parameter(s) belong.
            Must only be provided if parameter is specific to a certain coefficient.
        reaction_indices : int or list of int, optional
            The index(es) of the reaction(s) in the associated model(s), if applicable.
            Must only be provided if parameter is specific to a certain reaction.
        bound_state_indices : int or list of int, optional
            The index(es) of the bound state(s) in the associated model(s), if applicable.
            Must only be provided if parameter is specific to a certain bound state.
        section_indices : int or list of int, optional
            The index(es) of the section(s) in the associated model(s), if applicable.
            Must only be provided if parameter is specific to a certain section.
        abstols : float or list of float, optional
            The absolute tolerances for each parameter.
            If not provided, a default tolerance will be used.
        factors : float or list of float, optional
            The factors for each parameter.
            If not provided, a default factor of 1 will be used.

        Raises
        ------
        CADETProcessError
            Number of indices do not match for:
            - components
            - polynomial_coefficients
            - reaction
            - bound_state
            - sections
            - tolerances
            - factors

            Component is not found.
            Unit is not found.
            Parameter is not found.
            Name is not provided (if number of parameters larger than 1).
            If sensitivity name already exists.

        Notes
        -----
        This functionality is still work in progress.

        .. todo::
            - [ ] Check if compoment/reaction/polynomial index are required.
            - [ ] Specify time instead of section index;
        """
        if not isinstance(parameter_paths, list):
            parameter_paths = [parameter_paths]
        n_params = len(parameter_paths)

        if name is None:
            if n_params > 1:
                raise CADETProcessError(
                    "Must provide sensitivity name if n_params > 1."
                )
            else:
                name = parameter_paths[0]

        if name in self.parameter_sensitivity_names:
            raise CADETProcessError(
                "Parameter sensitivity with same name already exists."
            )

        if components is None:
            components = n_params * [None]
        if not isinstance(components, list):
            components = [components]
        if len(components) != n_params:
            raise CADETProcessError("Number of component indices does not match.")
        if components is None:
            components = n_params * [None]

        if not isinstance(polynomial_coefficients, list):
            polynomial_coefficients = [polynomial_coefficients]
        if len(polynomial_coefficients) != n_params:
            raise CADETProcessError("Number of coefficient indices does not match.")

        if reaction_indices is None:
            reaction_indices = n_params * [None]
        if not isinstance(reaction_indices, list):
            reaction_indices = [reaction_indices]
        if len(reaction_indices) != n_params:
            raise CADETProcessError("Number of reaction indices does not match.")

        if bound_state_indices is None:
            bound_state_indices = n_params * [None]
        if not isinstance(bound_state_indices, list):
            bound_state_indices = [bound_state_indices]
        if len(bound_state_indices) != n_params:
            raise CADETProcessError("Number of bound_state indices does not match.")

        if section_indices is None:
            section_indices = n_params * [None]
        if not isinstance(section_indices, list):
            section_indices = [section_indices]
        if len(section_indices) != n_params:
            raise CADETProcessError("Number of section indices does not match.")

        if abstols is None:
            abstols = n_params * [None]
        if not isinstance(abstols, list):
            abstols = [abstols]
        if len(abstols) != n_params:
            raise CADETProcessError("Number of abstol entries does not match.")

        if factors is None:
            factors = n_params * [1]
        if not isinstance(factors, list):
            factors = [factors]
        if len(factors) != n_params:
            raise CADETProcessError("Number of factor entries does not match.")

        units = []
        associated_models = []
        parameters = []
        for param, comp, coeff, reac, state, section, tol, fac in zip(
            parameter_paths,
            components,
            polynomial_coefficients,
            reaction_indices,
            bound_state_indices,
            section_indices,
            abstols,
            factors,
        ):
            param_parts = param.split(".")
            unit = param_parts[0]
            parameter = param_parts[-1]
            if parameter == "flow_rate":
                raise CADETProcessError(
                    "Flow rate is currently not supported for sensitivities."
                )
            parameters.append(parameter)

            associated_model = None
            if len(param_parts) == 3:
                associated_model = param_parts[1]

            if comp is not None and comp not in self.component_system.species:
                raise CADETProcessError(f"Unknown component {comp}.")

            unit = self.flow_sheet[unit]
            if unit not in self.flow_sheet.units:
                raise CADETProcessError("Not a valid unit")
            units.append(unit)

            if coeff is not None and parameter not in unit.polynomial_parameters:
                raise CADETProcessError("Not a polynomial parameter.")
            if parameter in unit.polynomial_parameters and coeff is None:
                raise CADETProcessError("Polynomial coefficient must be provided.")

            if associated_model is None:
                if parameter not in unit.parameters:
                    raise CADETProcessError("Not a valid parameter.")
            else:
                associated_model = getattr(unit, associated_model)

                if state is not None and state > associated_model.n_binding_sites:
                    raise ValueError("Binding site index exceed number of binding sites.")
                if reac is not None and reac > associated_model.n_reactions:
                    raise ValueError("Reaction index exceed number of reactions.")

                if parameter not in associated_model.parameters:
                    raise CADETProcessError("Not a valid parameter")
            associated_models.append(associated_model)

        sens = ParameterSensitivity(
            name,
            units,
            parameters,
            associated_models,
            components,
            polynomial_coefficients,
            reaction_indices,
            bound_state_indices,
            section_indices,
            abstols,
            factors,
        )
        self._parameter_sensitivities.append(sens)

    @property
    def system_state(self) -> np.ndarray:
        """np.ndarray: State of the entire system."""
        return self._system_state

    @system_state.setter
    def system_state(self, system_state: np.ndarray) -> None:
        self._system_state = system_state

    @property
    def system_state_derivative(self) -> np.ndarray:
        """np.ndarray: State derivative of the entire system."""
        return self._system_state_derivative

    @system_state_derivative.setter
    def system_state_derivative(self, system_state_derivative: np.ndarray) -> None:
        self._system_state_derivative = system_state_derivative

    @property
    def parameters(self) -> dict:
        """dict: Parameters of the process."""
        parameters = super().parameters

        parameters["flow_sheet"] = self.flow_sheet.parameters

        return parameters

    @parameters.setter
    def parameters(self, parameters: dict) -> None:
        try:
            self.flow_sheet.parameters = parameters.pop("flow_sheet")
        except KeyError:
            pass

        super(Process, self.__class__).parameters.fset(self, parameters)

    @property
    def section_dependent_parameters(self) -> Dict:
        """dict: Section dependent parameters of the process."""
        parameters = Dict()
        parameters.flow_sheet = self.flow_sheet.section_dependent_parameters
        return parameters

    @property
    def polynomial_parameters(self) -> Dict:
        """dict: Polynomial parameters of the process."""
        parameters = super().polynomial_parameters
        parameters.flow_sheet = self.flow_sheet.polynomial_parameters
        return parameters

    @property
    def sized_parameters(self) -> Dict:
        """dict: Sized parameters of the process."""
        parameters = super().sized_parameters
        parameters.flow_sheet = self.flow_sheet.sized_parameters
        return parameters

    @property
    def initial_state(self) -> dict:
        """dict: Initial state of the process."""
        initial_state = {state: getattr(self, state) for state in self._initial_states}
        initial_state["flow_sheet"] = self.flow_sheet.initial_state

        return initial_state

    @initial_state.setter
    def initial_state(self, initial_state: dict) -> None:
        try:
            self.flow_sheet.initial_state = initial_state.pop("flow_sheet")
        except KeyError:
            pass

        for state_name, state_value in initial_state.items():
            if state_name not in self._initial_state:
                raise CADETProcessError("Not an valid state")
            setattr(self, state_name, state_value)

    @property
    def config(self) -> Dict:
        """dict[str, dict]: Parameters and initial state of the process."""
        return Dict(
            {"parameters": self.parameters, "initial_state": self.initial_state}
        )

    @config.setter
    def config(self, config: dict) -> None:
        self.parameters = config["parameters"]
        self.initial_state = config["initial_state"]

    def add_concentration_profile(
        self,
        unit: str,
        time: np.ndarray,
        c: np.ndarray,
        components: Optional[list[str]] = None,
        s: float = 1e-6,
        interpolation_method: Literal["cubic", "pchip", None] = "pchip",
    ) -> None:
        """
        Add concentration profile to Process.

        Parameters
        ----------
        unit : str
            The name of the inlet unit operation.
        time : np.ndarray
            1D array containing the time values of the concentration profile.
        c : np.ndarray
            2D array containing the concentration profile with shape
            (len(time), n_comp), where n_comp is the number of components
            specified in the `components` argument.
        components : list[str] | None, optional
            Component species for which the concentration profile shall be added.
            If `None`, the profile is expected to have shape (len(time), n_comp).
            If `-1`, the same (1D) profile is added to all components.
            Default is `None`.
        s : float, optional
            Smoothing factor used to generate the spline representation of the
            concentration profile. Default is `1e-6`.
        interpolation_method : Literal["linear", "cubic", "pchip", None], optional
            The interpolation method to use. Options:
            - `"cubic"` : Cubic spline interpolation.
            - `"pchip"` : Piecewise cubic Hermite interpolation (default).
            - `None` : No interpolation, use raw time data.

        Raises
        ------
        TypeError
            If the specified `unit` is not an Inlet unit operation.
        ValueError
            If the time values in `time` exceed the cycle time of the Process.
            If `c` has an invalid shape.
            If `interpolation_method` is unknown.
        """
        if isinstance(unit, str):
            unit = self.flow_sheet[unit]

        if unit not in self.flow_sheet.inlets:
            raise TypeError("Expected Inlet")

        if max(time) > self.cycle_time:
            raise ValueError("Inlet profile exceeds cycle time.")

        # Handle components and concentration shape
        if components is None:
            if c.shape[1] != self.n_comp:
                raise ValueError(
                    f"Expected shape ({len(time), self.n_comp}) for concentration array"
                    f". Got {c.shape}."
                )
            components = self.component_system.species
        elif components == -1:
            # Assume same profile for all components
            if c.ndim > 1:
                raise ValueError("Expected single concentration profile.")
            c = np.column_stack([c] * self.n_comp)
            components = self.component_system.species

        if not isinstance(components, list):
            components = [components]

        indices = [self.component_system.species_indices[comp] for comp in components]
        if len(indices) == 1 and c.ndim == 1:
            c = np.array(c, ndmin=2).T

        # Determine the interpolation method
        if interpolation_method is None:
            time_interp = time
        else:
            time_interp = np.linspace(0, max(time), max(1001, len(time)))

            match interpolation_method:
                case None:
                    pass
                case "cubic":
                    interpolator = interpolate.CubicSpline(time, c)
                case "pchip":
                    interpolator = interpolate.PchipInterpolator(time, c)
                case _:
                    raise ValueError(
                        f"Unknown `interpolation_method`: {interpolation_method}."
                    )

            c = interpolator(time_interp)

        # Process each component's profile
        for i, comp in enumerate(indices):
            profile = c[:, i]

            # Normalize the profile
            min_val = np.min(profile)
            max_val = np.max(profile)
            range_val = max_val - min_val

            if range_val == 0:
                warn(
                    f"Component {comp} has no variation in concentration; "
                    "scaling will be skipped."
                )
                range_val = 1  # Avoid division by zero

            normalized_profile = (profile - min_val) / range_val

            # Fit the spline on the normalized profile
            tck = interpolate.splrep(time_interp, normalized_profile, s=s)
            ppoly = interpolate.PPoly.from_spline(tck)

            # Add events with properly unscaled coefficients
            for j, (t, sec) in enumerate(zip(ppoly.x, ppoly.c.T)):
                if j < 3 or j > len(ppoly.x) - 5:
                    continue
                # Scale coefficients and adjust the constant term
                scaled_sec = sec * range_val
                scaled_sec[-1] += min_val  # Adjust the constant term for shifting
                self.add_event(
                    f"{unit}_inlet_{comp}_{j - 3}",
                    f"flow_sheet.{unit}.c",
                    np.flip(scaled_sec),
                    t,
                    comp,
                )

    def add_flow_rate_profile(
        self,
        unit: str,
        time: np.ndarray,
        flow_rate: np.ndarray,
        s: float = 1e-6,
        interpolation_method: Literal["cubic", "pchip", None] = "pchip",
    ) -> None:
        """
        Add flow rate profile to a SourceMixin unit operation.

        Parameters
        ----------
        unit : str
            The name of the SourceMixin unit operation.
        time : np.ndarray
            1D array containing the time values of the flow rate profile.
        flow_rate : np.ndarray
            1D array containing the flow rate values over time.
        s : float, optional
            Smoothing factor used to generate the spline representation of the flow rate profile.
            Default is `1e-6`.
        interpolation_method : Literal["linear", "cubic", "pchip", None], optional
            The interpolation method to use. Options:
            - `"cubic"` : Cubic spline interpolation.
            - `"pchip"` : Piecewise cubic Hermite interpolation (default).
            - `None` : No interpolation, use raw time data.

        Raises
        ------
        TypeError
            If the specified `unit` is not a SourceMixin unit operation.
        ValueError
            If the time values in `time` exceed the cycle time of the Process.
        """
        if isinstance(unit, str):
            unit = self.flow_sheet[unit]

        if unit not in self.flow_sheet.inlets + self.flow_sheet.cstrs:
            raise TypeError("Expected SourceMixin.")

        if max(time) > self.cycle_time:
            raise ValueError("Inlet profile exceeds cycle time.")

        # Compute min and max for scaling
        min_val = np.min(flow_rate)
        max_val = np.max(flow_rate)
        range_val = max_val - min_val

        if range_val == 0:
            warn("Flow rate has no variation; scaling will be skipped.")
            range_val = 1  # Avoid division by zero, effectively skipping scaling

        # Normalize flow_rate to [0, 1]
        normalized_flow_rate = (flow_rate - min_val) / range_val

        # Determine interpolation method
        if interpolation_method is None:
            time_interp = time
            interpolated_flow_rate = normalized_flow_rate
        else:
            time_interp = np.linspace(0, max(time), max(1001, len(time)))
            match interpolation_method:
                case "cubic":
                    interpolator = interpolate.CubicSpline(time, normalized_flow_rate)
                case "pchip":
                    interpolator = interpolate.PchipInterpolator(
                        time, normalized_flow_rate
                    )
                case _:
                    raise ValueError(
                        f"Unknown `interpolation_method`: {interpolation_method}."
                    )

            interpolated_flow_rate = interpolator(time_interp)

        # Fit the spline with the interpolated normalized flow rate
        tck = interpolate.splrep(time_interp, interpolated_flow_rate, s=s)
        ppoly = interpolate.PPoly.from_spline(tck)

        # Add events with unscaled coefficients
        for i, (t, sec) in enumerate(zip(ppoly.x, ppoly.c.T)):
            if i < 3 or i > len(ppoly.x) - 5:
                continue

            # Unscale all coefficients
            unscaled_sec = sec * range_val
            unscaled_sec[-1] += min_val  # Adjust the constant term for shifting

            self.add_event(
                f"{unit}_flow_rate_{i - 3}",
                f"flow_sheet.{unit}.flow_rate",
                np.flip(unscaled_sec),
                t,
            )

    def check_config(self) -> bool:
        """
        Validate that process config is setup correctly.

        Returns
        -------
        check : Bool
            True if process is setup correctly. False otherwise.
        """
        flag = super().check_config()

        missing_parameters = self.flow_sheet.missing_parameters
        if len(missing_parameters) > 0:
            for param in missing_parameters:
                if f"flow_sheet.{param}" not in self.event_parameters:
                    warn(f"Missing parameter {param}.")
                    flag = False

        if not self.flow_sheet.check_connections():
            flag = False

        if self.cycle_time is None:
            warn("Cycle time is not set")
            flag = False

        if not self.check_cstr_volume():
            flag = False

        return flag

    def check_cstr_volume(self) -> bool:
        """
        Check if CSTRs run empty.

        Returns
        -------
        flag : bool
            False if any of the CSTRs run empty. True otherwise.
        """
        flag = True
        for cstr in self.flow_sheet.cstrs:
            if cstr.flow_rate is None:
                continue
            V_0 = cstr.init_liquid_volume
            unit_index = self.flow_sheet.get_unit_index(cstr)
            for port in self.flow_sheet.units[unit_index].ports:
                V_in = self.flow_rate_timelines[cstr.name].total_in[port].integral()
                V_out = self.flow_rate_timelines[cstr.name].total_out[port].integral()
                if V_0 + V_in - V_out < 0:
                    flag = False
                    warn(f"CSTR {cstr.name} runs empty on port {port} during process.")

        return flag

    def __str__(self) -> str:
        """str: String representation of the process."""
        return self.name


@dataclass
class ParameterSensitivity:
    """Class for storing parameter sensitivity parameters."""

    name: str
    units: list
    parameters: list
    associated_models: list = None
    components: list = None
    polynomial_coefficients: list = None
    reaction_indices: list = None
    bound_state_indices: list = None
    section_indices: list = None
    abstols: list = None
    factors: list = None



"""
=========================================
Smoothing (:mod:`CADETProcess.smoothing`)
=========================================

.. currentmodule:: CADETProcess.smoothing

This module provides functionality for smoothing data.


.. autosummary::
    :toctree: generated/

    find_smoothing_factors
    full_smooth

"""  # noqa

import multiprocessing
from typing import Any, Callable, Optional

import numpy as np
import numpy.typing as npt
import scipy.signal
from pymoo.algorithms.soo.nonconvex.pattern import PatternSearch
from pymoo.core.problem import ElementwiseProblem
from pymoo.optimize import minimize
from scipy.interpolate import UnivariateSpline

butter_order = 3

__all__ = ["find_smoothing_factors", "full_smooth"]


class TargetProblem(ElementwiseProblem):
    def __init__(
        self,
        lb: float,
        ub: float,
        sse_target: float,
        func: Callable,
        values: npt.ArrayLike,
        fs: float,
    ) -> None:
        super().__init__(n_var=1, n_obj=1, n_constr=0, xl=lb, xu=ub)
        self.sse_target = sse_target
        self.func = func
        self.values = values
        self.fs = fs

    def _evaluate(
        self,
        crit_fs: float,
        out: dict,
        *args: Optional[tuple],
        **kwargs: Optional[dict],
    ) -> None:
        crit_fs = 10**crit_fs
        try:
            sos = self.func(crit_fs, self.fs)
            low_passed = scipy.signal.sosfiltfilt(sos, self.values)
            sse = np.sum((low_passed - self.values) ** 2)

            error = (sse - self.sse_target) ** 2
        except (ValueError, np.linalg.LinAlgError):
            error = np.inf
        out["F"] = error


class MaxDistance(ElementwiseProblem):
    def __init__(
        self,
        lb: float,
        ub: float,
        func: Callable,
        fs: float,
        values: npt.ArrayLike,
        x_min: float,
        y_min: float,
        p1: float,
        p2: float,
        factor: float,
    ) -> None:
        super().__init__(n_var=1, n_obj=1, n_constr=0, xl=lb, xu=ub)
        self.func = func
        self.fs = fs
        self.values = values
        self.x_min = x_min
        self.y_min = y_min
        self.p1 = p1
        self.p2 = p2
        self.factor = factor

    def _evaluate(
        self,
        crit_fs: float,
        out: dict,
        *args: Optional[tuple],
        **kwargs: Optional[dict],
    ) -> None:
        crit_fs = 10.0 ** crit_fs[0]
        try:
            sos = self.func(crit_fs, self.fs)
        except ValueError:
            out["F"] = 1e6
            return

        try:
            low_passed = scipy.signal.sosfiltfilt(sos, self.values)
        except np.linalg.LinAlgError:
            out["F"] = 1e6
            return

        sse = np.sum((low_passed - self.values) ** 2)

        pT = np.array([crit_fs - self.x_min, np.log(sse) - self.y_min]).T / self.factor

        d = np.cross(self.p2 - self.p1, self.p1 - pT) / np.linalg.norm(
            self.p2 - self.p1
        )
        out["F"] = -d


def get_p(x: npt.ArrayLike, y: npt.ArrayLike) -> tuple:
    x = np.array(x)
    y = np.array(y)

    sort_idx = np.argsort(x)

    x = x[sort_idx]
    y = y[sort_idx]

    y_min = y - min(y)
    x_min = x - min(x)

    p3 = np.array([x_min, y_min]).T
    factor = np.max(p3, 0)
    p3 = p3 / factor
    p1 = p3[0, :]
    p2 = p3[-1, :]

    return x, min(x), y, min(y), p1, p2, p3, factor


def signal_bessel(crit_fs: float, fs: float) -> np.ndarray:
    return scipy.signal.bessel(
        butter_order,
        crit_fs,
        btype="lowpass",
        analog=False,
        fs=fs,
        output="sos",
        norm="delay",
    )


def signal_butter(crit_fs: float, fs: float) -> np.ndarray:
    return scipy.signal.butter(
        butter_order, crit_fs, btype="lowpass", analog=False, fs=fs, output="sos"
    )


def refine_signal(
    func: Callable,
    times: npt.ArrayLike,
    values: npt.ArrayLike,
    x: npt.ArrayLike,
    y: npt.ArrayLike,
    fs: float,
    start: float,
) -> float:
    x, x_min, y, y_min, p1, p2, p3, factor = get_p(x, y)

    lb = np.log10(x[0])
    ub = np.log10(x[-1])

    problem = MaxDistance(lb, ub, func, fs, values, x_min, y_min, p1, p2, factor)

    algorithm = PatternSearch(n_sample_points=50, eps=1e-13)

    res = minimize(problem, algorithm, verbose=False, seed=1)

    crit_fs = 10 ** res.X[0]

    return crit_fs


def find_L(x: npt.ArrayLike, y: npt.ArrayLike) -> tuple[float, float]:
    # find the largest value greater than 0
    # otherwise return none to just turn off butter filter
    x, x_min, y, y_min, p1, p2, p3, factor = get_p(x, y)

    d = np.cross(p2 - p1, p1 - p3) / np.linalg.norm(p2 - p1)

    max_idx = np.argmax(d)
    max_d = d[max_idx]
    l_x = x[max_idx]
    l_y = y[max_idx]

    if max_d <= 0:
        return None, None

    return l_x, l_y


def find_signal(
    func: Callable,
    times: npt.ArrayLike,
    values: npt.ArrayLike,
    sse_target: float,
) -> float:
    filters = []
    sse = []

    fs = 1.0 / (times[1] - times[0])

    ub = fs / 2.0

    for i in np.logspace(-8, np.log10(ub), 50):
        try:
            sos = func(i, fs)
            low_passed = scipy.signal.sosfiltfilt(sos, values)

            filters.append(i)
            sse.append(np.sum((low_passed - values) ** 2))
        except (ValueError, np.linalg.LinAlgError):
            continue

    crit_fs_max = find_max_signal(func, times, values, sse_target, filters, sse)

    L_x, L_y = find_L(filters, np.log(sse))

    if L_x is not None:
        L_x = refine_signal(func, times, values, filters, np.log(sse), fs, L_x)

    if L_x is not None and crit_fs_max < L_x:
        L_x = crit_fs_max

    return L_x


def find_max_signal(
    func: Callable,
    times: npt.ArrayLike,
    values: npt.ArrayLike,
    sse_target: float,
    filters: list,
    sse: Any,
) -> float:
    fs = 1.0 / (times[1] - times[0])

    filters = np.log10(filters)
    problem = TargetProblem(filters[0], filters[-1], sse_target, func, values, fs)

    algorithm = PatternSearch(n_sample_points=50, eps=1e-13)

    res = minimize(problem, algorithm, verbose=False, seed=1)

    crit_fs = 10 ** res.X[0]

    return crit_fs


def smoothing_filter_signal(
    func: Callable,
    times: npt.ArrayLike,
    values: npt.ArrayLike,
    crit_fs: float,
) -> np.ndarray:
    if crit_fs is None:
        return values
    fs = 1.0 / (times[1] - times[0])
    sos = func(crit_fs, fs)
    low_passed = scipy.signal.sosfiltfilt(sos, values)
    return low_passed


def find_smoothing_factors(
    times: npt.ArrayLike,
    values: npt.ArrayLike,
    rmse_target: float = 1e-4,
) -> tuple:
    """Find smoothing factors."""
    sse_target = (rmse_target**2.0) * len(values)

    try:
        crit_fs = find_signal(signal_bessel, times, values, sse_target)
    except np.linalg.LinAlgError:
        crit_fs = None

    if crit_fs is None:
        multiprocessing.get_logger().info(
            "butter filter disabled, no viable L point found"
        )

    values_filter = smoothing_filter_signal(signal_bessel, times, values, crit_fs)

    s = sse_target

    spline, factor = create_spline(times, values, crit_fs, s)

    # run a quick butter pass to remove high frequency noise in the derivative
    # (needed for some experimental data)
    values_filter = spline.derivative()(times) / factor
    factor = 1.0 / np.max(values_filter)
    values_filter = values_filter * factor
    crit_fs_der = find_signal(signal_bessel, times, values_filter, sse_target)

    return s, crit_fs, crit_fs_der


def create_spline(
    times: npt.ArrayLike,
    values: npt.ArrayLike,
    crit_fs: float,
    s: Optional[float],
) -> tuple[UnivariateSpline, float]:
    factor = 1.0 / np.max(values)
    values = values * factor
    values_filter = smoothing_filter_signal(signal_bessel, times, values, crit_fs)

    return (
        scipy.interpolate.UnivariateSpline(times, values_filter, s=s, k=5, ext=3),
        factor,
    )


def smooth_data(
    times: npt.ArrayLike,
    values: npt.ArrayLike,
    crit_fs: float,
    s: float,
) -> float:
    spline, factor = create_spline(times, values, crit_fs, s)

    return spline(times) / factor


def smooth_data_derivative(
    times: npt.ArrayLike,
    values: npt.ArrayLike,
    crit_fs: float,
    s: float,
    crit_fs_der: float,
    smooth: bool = True,
) -> float:
    spline, factor = create_spline(times, values, crit_fs, s)

    if smooth:
        values_filter_der = spline.derivative()(times) / factor

        factor_der = 1.0 / np.max(values_filter_der)
        values_filter_der = values_filter_der * factor_der
        values_filter_der = butter(times, values_filter_der, crit_fs_der)
        values_filter_der = values_filter_der / factor_der
        spline_der = scipy.interpolate.InterpolatedUnivariateSpline(
            times, values_filter_der, k=5, ext=3
        )
        values_filter_der = spline_der(times)
    else:
        values_filter_der = spline.derivative()(times) / factor
    return values_filter_der


def full_smooth(
    times: npt.ArrayLike,
    values: npt.ArrayLike,
    crit_fs: float,
    s: float,
    crit_fs_der: float,
    smooth: bool = True,
) -> tuple:
    """Create full smooth data."""
    spline, factor = create_spline(times, values, crit_fs, s)

    values_filter = spline(times) / factor

    # run a quick butter pass to remove high frequency noise in the derivative
    # (needed for some experimental data)

    if smooth:
        values_filter_der = spline.derivative()(times) / factor

        factor_der = 1.0 / np.max(values_filter_der)
        values_filter_der = values_filter_der * factor_der
        values_filter_der = butter(times, values_filter_der, crit_fs_der)
        values_filter_der = values_filter_der / factor_der
        spline_der = scipy.interpolate.InterpolatedUnivariateSpline(
            times, values_filter_der, k=5, ext=3
        )
        values_filter_der = spline_der(times)
    else:
        values_filter_der = spline.derivative()(times) / factor
    return values_filter, values_filter_der


def butter(
    times: npt.ArrayLike,
    values: npt.ArrayLike,
    crit_fs_der: float,
) -> np.ndarray:
    values_filter = smoothing_filter_signal(signal_bessel, times, values, crit_fs_der)

    return values_filter



from __future__ import annotations

import itertools
import warnings
from typing import Optional

import numpy as np
import numpy.typing as npt
import scipy
from matplotlib.axes import Axes
from numpy.exceptions import VisibleDeprecationWarning

from CADETProcess import CADETProcessError, plotting
from CADETProcess.dataStructure import NdPolynomial, Structure

__all__ = ["Section", "TimeLine", "MultiTimeLine"]


class Section(Structure):
    """
    Helper class to store parameter states between events.

    Attributes
    ----------
    start : float
        Start time of section
    end : float
        End time of section.
    coeffs : int or float or array_like
        Polynomial coefficients of state in order of increasing degree.
    n_entries : int
        Number of entries (e.g. components, output_states)
    degree : int
        Degree of polynomial to represent state.

    Notes
    -----
        if coeffs is int: Set constant value for for all entries
        if coeffs is list: Set value per component (check length!)
        if coeffs is ndarray (or list of lists): set polynomial coefficients
    """

    coeffs = NdPolynomial(size=("n_entries", "n_poly_coeffs"))

    def __init__(
        self,
        start: float,
        end: float,
        coeffs: int | float | npt.ArrayLike,
        is_polynomial: bool = False,
    ) -> None:
        """Construct section object."""
        if start > end:
            raise ValueError("End time must be greater than start time")

        self.start = start
        self.end = end
        diff = end - start

        coeffs = np.array(coeffs, ndmin=1, dtype=np.float64)
        self.parameter_shape = coeffs.shape

        if is_polynomial:
            coeffs = np.array(coeffs, ndmin=2, dtype=np.float64)
            self.degree = coeffs.shape[-1] - 1
            self.n_entries = coeffs.shape[0]

            self.coeffs = coeffs
        else:
            self.degree = 0
            self.n_entries = coeffs.size
            self.coeffs = coeffs.reshape((coeffs.size, 1))

        self._poly = []
        for i in range(self.n_entries):
            poly = np.polynomial.polynomial.Polynomial(
                self.coeffs[i], domain=(start, end), window=(0, diff)
            )
            self._poly.append(poly)

        self._poly_der = []
        for iEntry in range(self.n_entries):
            with warnings.catch_warnings():
                warnings.simplefilter("ignore")

                poly_der = self._poly[iEntry].deriv(1)
                self._poly_der.append(poly_der)

    @property
    def is_polynomial(self) -> bool:
        """bool: True if Section represents polynomial parameter. False otherwise."""
        if self.degree > 0:
            return True
        return False

    @property
    def n_poly_coeffs(self) -> int:
        """int: Number of polynomial coefficients."""
        return self.degree + 1

    @property
    def is_single_entry(self) -> bool:
        """bool: True if Section contains single entry. False otherwise."""
        if self.n_entries > 1:
            return True

        if self.is_polynomial and self.parameter_shape.ndim == 1:
            return True

        return False

    def value(self, t: float) -> float:
        """
        Return value of parameter section at time t.

        Parameters
        ----------
        t : float
            Time at which function is evaluated.

        Returns
        -------
        y : float
            Value of parameter state at time t.

        Raises
        ------
        ValueError
            If t is lower than start or larger than end of section time.
        """
        if np.any(t < self.start) or np.any(self.end < t):
            raise ValueError("Time exceeds section times")

        value = np.array([p(t) for p in self._poly])

        return value

    __call__ = value

    def coefficients(self, offset: float = 0.0) -> np.ndarray:
        """
        Get coefficients at (time) offset.

        Parameters
        ----------
        offset : float
            (Time) offset to be evaluated.

        Returns
        -------
        coeffs : np.ndarray
            Coefficients at offset.
        """
        coeffs = []
        for i in range(self.n_entries):
            c = self.coeffs[i].copy()
            c[0] = self._poly[i](offset)
            if self.degree > 0:
                c[1] = self._poly_der[i](offset)
            coeffs.append(c)

        return np.array(coeffs).reshape(self.parameter_shape)

    def derivative(self, t: float, order: Optional[int] = 1) -> np.ndarray:
        """
        Return derivative of parameter section at time t.

        Parameters
        ----------
        t : float
            Time at which function is evaluated.
        order : int, default=1
            Order of deriviation. @TODO: Not yet implemented.

        Returns
        -------
        y_dot : float
            Derivative of parameter state at time t.

        Raises
        ------
        ValueError
            If t is lower than start or larger than end of section time.
        ValueError
            If order is larger than polynomial degree
        """
        if np.any(t < self.start) or np.any(self.end < t):
            raise ValueError("Time exceeds section times")

        deriv = np.array([p.deriv(t).coef for p in self._poly_der])

        return deriv

    def integral(
        self, start: Optional[float] = None, end: Optional[float] = None
    ) -> np.ndarray:
        """
        Return integral of function in interval [start, end].

        Parameters
        ----------
        start : float, optional
            Lower integration bound.
        end : float, optional
            Upper integration bound.

        Returns
        -------
        Y : np.ndarray
            Value of definite integral between start and end.

        Raises
        ------
        ValueError
            If integration bounds exceed section times.
        """
        if start is None:
            start = self.start
        if end is None:
            end = self.end

        if not ((self.start <= start) & (start <= end) & (end <= self.end)):
            raise ValueError("Integration bounds exceed section times")

        integ_methods = [p.integ(lbnd=start) for p in self._poly]
        return np.array([i(end) for i in integ_methods])

    def __repr__(self) -> str:
        """str: String representation of the Section."""
        args = f"start={self.start}, end={self.end}, coeffs={self.coeffs}"
        if self.degree > 0:
            args += f", degree={self.degree}"

        return f"Section({args})"


class TimeLine:
    """
    Class representing a timeline of time-varying data.

    The timeline is made up of Sections, which are continuous time intervals.
    Each Section represents a piecewise polynomial function that defines the
    variation of a given parameter over time.

    Attributes
    ----------
    sections : List[Section]
        List of Sections that make up the timeline.
    """

    def __init__(self) -> None:
        """Initialize TimeLine object."""
        self._sections = []

    @property
    def sections(self) -> list:
        """list: Sections of the TimeLine."""
        return self._sections

    @property
    def degree(self) -> int:
        """int: Degree of the polynomial functions used to represent each Section."""
        if len(self.sections) > 0:
            return self.sections[0].degree

    @property
    def n_entries(self) -> int:
        """int: Number of entries in the parameter vector for each Section."""
        if len(self.sections) > 0:
            return self.sections[0].n_entries

    def add_section(self, section: Section) -> None:
        """
        Add a Section to the timeline.

        Parameters
        ----------
        section : Section
            The Section to be added to the timeline.

        Raises
        ------
        TypeError
            If section is not an instance of the Section class.
        CADETProcessError
            If the polynomial degree of the Section does not match the degree of
            the other Sections in the timeline.
        CADETProcessError
            If the Section introduces a gap in the timeline.
        """
        if not isinstance(section, Section):
            raise TypeError("Expected Section")
        if len(self.sections) > 0:
            if section.degree != self.degree:
                raise CADETProcessError("Polynomial degree does not match")

            if not (section.start == self.end or section.end == self.start):
                raise CADETProcessError("Sections times must be without gaps")

        self._sections.append(section)
        self._sections = sorted(self._sections, key=lambda sec: sec.start)

        self.update_piecewise_poly()

    def update_piecewise_poly(self) -> None:
        """Update the piecewise polynomial representation of the timeline."""
        x = []
        coeffs = []
        for sec in self.sections:
            coeffs.append(np.array(sec.coeffs))
            x.append(sec.start)
        x.append(sec.end)

        piecewise_poly = []
        for iEntry in range(self.n_entries):
            c = np.array([iCoeff[iEntry, :] for iCoeff in coeffs])
            c_decreasing = np.fliplr(c)
            p = scipy.interpolate.PPoly(c_decreasing.T, x)
            piecewise_poly.append(p)

        self._piecewise_poly = piecewise_poly

    @property
    def piecewise_poly(self) -> list:
        """list: scipy.interpolate.PPoly for each dimension."""
        return self._piecewise_poly

    def value(self, time: float) -> np.ndarray:
        """
        np.ndarray: Value of parameter at given time.

        Parameters
        ----------
        time : np.float or array_like
            time points at which to evaluate.
        """
        return np.array([p(time) for p in self.piecewise_poly]).T

    def coefficients(self, time: float) -> np.ndarray:
        """
        Return coefficient of polynomial at given time.

        Parameters
        ----------
        time : float
            Time at which polynomial coefficients are queried.

        Returns
        -------
        coefficients : np.ndarray
            !!! Array of coefficients in ORDER !!!
        """
        section_index = self.section_index(time)
        c = self.sections[section_index].coefficients(time)

        return c

    def integral(
        self, start: Optional[float] = None, end: Optional[float] = None
    ) -> np.ndarray:
        """
        Calculate integral of sections in interval [start, end].

        Parameters
        ----------
        start : float, optional
            Lower integration bound.
        end : float, optional
            Upper integration bound.

        Returns
        -------
        Y : np.ndarray
            Value of definite integral between start and end.

        Raises
        ------
        ValueError
            If integration bounds exceed section times.
        """
        if start is None:
            start = self.start
        if end is None:
            end = self.end

        if not ((self.start <= start) & (start <= end) & (end <= self.end)):
            raise ValueError("Integration bounds exceed section times")

        return np.array([p.integrate(start, end) for p in self.piecewise_poly]).T

    def section_index(self, time: float) -> int:
        """
        Return the index of the section that contains the specified time.

        Parameters
        ----------
        time : float
            The time to check.

        Returns
        -------
        int
            The index of the section that contains the specified time.
        """
        section_times = np.array(self.section_times)

        return np.argmin(time >= section_times) - 1

    @property
    def section_times(self) -> list[float]:
        """List of float: The start and end times of all sections in the timeline."""
        if len(self.sections) == 0:
            return []

        return [self.sections[0].start] + [sec.end for sec in self.sections]

    @property
    def start(self) -> float:
        """float: The start time of the timeline."""
        return self.section_times[0]

    @property
    def end(self) -> float:
        """float: The end time of the timeline."""
        return self.section_times[-1]

    @plotting.create_and_save_figure
    def plot(self, ax: Axes, x_axis_in_minutes: bool = True) -> Axes:
        """
        Plot the state of the timeline over time.

        Parameters
        ----------
        ax : Axes
            The axes to plot on.
        x_axis_in_minutes: bool, optional
            If True, the x-axis will be plotted using minutes. The default is True.

        Returns
        -------
        ax : Axes
            The axes with the plot of the timeline state.
        """
        start = self.sections[0].start
        end = self.sections[-1].end
        time = np.linspace(start, end, 1001)
        y = self.value(time)

        if x_axis_in_minutes:
            time = time / 60
            start = start / 60
            end = end / 60

        ax.plot(time, y)

        layout = plotting.Layout()
        layout.x_label = "$time~/~s$"
        if x_axis_in_minutes:
            layout.x_label = "$time~/~min$"
        layout.y_label = "$state$"
        layout.x_lim = (start, end)
        layout.y_lim = (np.min(y), 1.1 * np.max(y))

        plotting.set_layout(ax, layout)

        return ax

    @classmethod
    def from_constant(cls, start: float, end: float, value: float) -> TimeLine:
        """
        Create a timeline with a constant value for a given time range.

        Parameters
        ----------
        start : float
            The start time of the time range.
        end : float
            The end time of the time range.
        value : float
            The value of the timeline during the time range.

        Returns
        -------
        TimeLine
            A TimeLine instance with a single section with the specified constant value.
        """
        tl = cls()
        tl.add_section(Section(start, end, value))

        return tl

    @classmethod
    def from_profile(
        cls,
        time: npt.ArrayLike,
        profile: npt.ArrayLike,
        s: float = 1e-6,
    ) -> TimeLine:
        """
        Create a timeline from a profile.

        Parameters
        ----------
        time : array_like
            The time values of the profile.
        profile : array_like
            The profile values.
        smoothing : float, optional
            The smoothing factor for the spline interpolation. The default is 1e-6.

        Returns
        -------
        TimeLine
            A TimeLine instance with polynomial sections created from the profile.
        """
        from scipy import interpolate

        tl = cls()

        tck = interpolate.splrep(time, profile, s=s)
        ppoly = interpolate.PPoly.from_spline(tck)

        for i, (start, sec) in enumerate(zip(ppoly.x, ppoly.c.T)):
            if i < 3:
                continue
            elif i > len(ppoly.x) - 5:
                continue
            end = ppoly.x[i + 1]
            tl.add_section(Section(start, end, np.flip(sec), is_polynomial=True))

        return tl


class MultiTimeLine:
    """
    Class for a collection of TimeLines with the same number of entries.

    Attributes
    ----------
    base_state : np.ndarray
        The base state that each TimeLine represents.
    n_entries : int
        The number of entries in each TimeLine.
    time_lines : list of TimeLine
        The collection of TimeLines.
    degree : int
        The degree of the polynomials in each section.
    """

    def __init__(self, base_state: list, is_polynomial: bool = False) -> None:
        """
        Initialize a MultiTimeLine instance.

        Parameters
        ----------
        base_state : list
            The base state that each TimeLine represents.
        is_polynomial : bool, optional
            Option wheter MultiTimeLine is polynomial. The default is False

        """
        base_state = np.array(base_state, ndmin=1, dtype=np.float64)
        self.is_polynomial = is_polynomial

        if self.is_polynomial:
            if base_state.ndim == 1:
                self.is_single_entry = True
            else:
                self.is_single_entry = False
            self.base_state = np.array(base_state, ndmin=2, dtype=np.float64)
            self.degree = self.base_state.shape[-1] - 1
        else:
            self.degree = 0
            self.base_state = base_state

        self.time_lines = [TimeLine() for _ in range(self.size)]

    @property
    def degree(self) -> int:
        """int: The degree of the polynomials in each section."""
        return self._degree

    @degree.setter
    def degree(self, degree: int) -> None:
        self._degree = degree

    @property
    def n_entries(self) -> int:
        """int: Number of entries handled by MultiTimeline."""
        if self.degree > 0:
            return len(self.base_state)
        return self.base_state.size

    @property
    def size(self) -> int:
        """int: Total number of internal TimeLines handled Number by MultiTimeline."""
        return self.base_state.size

    @property
    def section_times(self) -> list:
        """list: Combined section times of all TimeLines."""
        time_line_sections = [tl.section_times for tl in self.time_lines]

        section_times = set(itertools.chain.from_iterable(time_line_sections))

        return sorted(list(section_times))

    def add_section(self, section: Section, entry_index: tuple) -> None:
        """
        Add section to TimeLine with specific entry index.

        Parameters
        ----------
        section : Section
            Section to be added.
        entry_index : tuple
            Index of the entry in the base_state for which the section will be added.

        Raises
        ------
        ValueError
            If entry index is out of bounds for base_state.
        """
        index = flatten_index(self.base_state.shape, entry_index)[0]
        if index > self.size:
            raise CADETProcessError("Entry index is out of bounds.")
        self.time_lines[index].add_section(section)

    @property
    def combined_time_line(self) -> TimeLine:
        """TimeLine: Object representing combination of all timelines in the MultiTimeLine."""
        tl = TimeLine()

        n_poly_coeffs = self.degree + 1

        coeffs = self.base_state

        section_times = self.section_times
        for iSec in range(len(section_times) - 1):
            start = self.section_times[iSec]
            end = self.section_times[iSec + 1]

            if not self.is_polynomial:
                for i, entry in enumerate(self.time_lines):
                    if len(entry.sections) > 0:
                        index = unflatten_index(self.base_state.shape, i)[0]
                        coeff = entry.coefficients(start)[0]
                        coeffs[index] = coeff
            else:
                for i_entry in range(self.n_entries):
                    tl_indices = slice(
                        i_entry * n_poly_coeffs, (i_entry + 1) * n_poly_coeffs
                    )
                    i_entry_tl = self.time_lines[tl_indices]
                    for i_poly, i_poly_tl in enumerate(i_entry_tl):
                        if self.is_single_entry:
                            index = (0, i_poly)
                        else:
                            index = (i_entry, i_poly)

                        if (
                            len(i_poly_tl.sections) > 0
                            and start in i_poly_tl.section_times
                        ):
                            coeffs[index] = i_poly_tl.coefficients(start)[0]

            section = Section(start, end, coeffs, self.is_polynomial)
            tl.add_section(section)

            coeffs = tl.coefficients(end)

        return tl


def generate_indices(
    shape: tuple[int, ...],
    indices: Optional[list[list[int]]] = None,
) -> list[tuple]:
    """
    Generate tuples representing indices for an array with a given shape.

    This method allows specifying a list of indices where each entry can also contain
    slices.

    Parameters
    ----------
    shape : tuple of int
        The shape of the array to be indexed.
    indices : list of list of int, optional
        A list where each sub-list contains indices for one dimension of the array.
        'None' indicates a full slice (':') for that dimension.

    Raises
    ------
    ValueError
        If 'parameter' is a scalar or if an index in 'indices' is out of bounds.

    Returns
    -------
    list
        A list of tuples, where each tuple represents indices into the 'parameter'
        array. If the 'parameter' array was 1D, this will be a list of integers instead.

    Examples
    --------
    >>> parameter = np.array([[1, 2, 3], [4, 5, 6], [7, 8, 9]])
    >>> indices = [[0, 1], [1, 2]]
    >>> generate_indices(parameter, indices)
    [(0, 1), (1, 2)]
    """
    if not shape:
        raise ValueError(
            "Shape must not be empty, scalar parameters are not supported."
        )

    if indices is None:
        indices = np.s_[:]

    if not isinstance(indices, list):
        indices = [
            indices,
        ]

    indices_array = np.array(indices, ndmin=1)

    indices = []
    for ind in indices_array:
        ind = tuple(np.array(ind, ndmin=1))
        indices.append(ind)

    _validate_indices(shape, indices)

    return indices


def _validate_indices(shape: tuple[int, ...], indices: list[list[int]]) -> None:
    """Validate that all indices can be set in an array with shape `shape`."""
    param_ref = np.arange(np.prod(shape)).reshape(shape)

    for ind in indices:
        param_ref[ind]


def unravel(shape: tuple[int, ...], indices: list[int] | tuple) -> list[tuple]:
    """
    Unravel indices of a multi-dimensional array.

    Parameters
    ----------
    shape : tuple of int
        The shape of the original array.
    indices : list of int or tuple
        The indices in the flattened array to be unraveled.

    Returns
    -------
    list of tuple
        The unraveled indices in the multi-dimensional array.
    """
    if len(shape) == 0:
        return []

    indices_flat_ref = np.arange(np.prod(shape)).reshape(shape)

    indices_unraveled = []
    for ind in indices:
        indices_flat = indices_flat_ref[ind].flatten()
        indices_unravel = np.unravel_index(indices_flat, shape)
        if not isinstance(indices_flat, (int, np.int64)):
            indices_unravel = list(zip(*indices_unravel))
        else:
            indices_unravel = [indices_unravel]
        indices_unraveled += indices_unravel

    return indices_unraveled


def flatten_index(shape: tuple[int], indices: tuple | list[tuple]) -> list[int]:
    """
    Flatten indices to access array.

    Parameters
    ----------
    shape: tuple
        Shape of the array to be indexed
    indices : tuple or list of tuples
        Indices in tuple notation

    Returns
    -------
    indices_flat : list
        List of indices in flat notation.
    """
    if not isinstance(indices, list):
        indices = [indices]

    indices_flat_ref = np.arange(np.prod(shape)).reshape(shape)

    return [indices_flat_ref[i] for i in indices]


def unflatten_index(shape: tuple[int, ...], indices_flat: int | list[int]) -> list[int]:
    """
    Unflatten indices to access array.

    Parameters
    ----------
    shape : tuple
        Shape of the array to be indexed
    indices_flat : int or list of ints
        Flat indices

    Returns
    -------
    indices : list
        List of unflattened indices.
    """
    if not isinstance(indices_flat, list):
        indices_flat = [indices_flat]

    indices_unravel = np.unravel_index(indices_flat, shape)
    indices = list(zip(*indices_unravel))

    return indices


def get_inhomogeneous_shape(value: np.ndarray) -> list[tuple[int, ...]]:
    """If array is inhomogeneous, return list with shape of every element."""
    with warnings.catch_warnings():  # Catch warnings for compatibility with numpy<1.24
        warnings.simplefilter("error")
        try:
            return np.array(value).shape
        except (ValueError, VisibleDeprecationWarning):
            pass

    shape = []

    for i in value:
        i_shape = get_inhomogeneous_shape(i)
        shape.append(i_shape)

    return shape


def get_full_shape(inhomogeneous_shape: list) -> tuple[int]:
    """Create full shape from inhomogeneous shape to be used with numpy arrays."""
    first_dimension = len(inhomogeneous_shape)

    sub_dims = ()
    for sub_dim in inhomogeneous_shape:
        if not isinstance(sub_dim, tuple):
            sub_dim = get_full_shape(sub_dim)

        sub_dims += (sub_dim,)

    max_dims = {}
    for el in sub_dims:
        for i, dim in enumerate(el):
            try:
                max_dims[i] = max(max_dims[i], dim)
            except KeyError:
                max_dims[i] = dim

    dims = (first_dimension,) + tuple(max_dims.values())

    return dims


def extract_inhomogeneous_array(
    full_array: np.ndarray, inhomogeneous_shape: tuple[int, ...]
) -> None:
    """Get inhomogeneous array from full_array."""
    array = []
    for i in inhomogeneous_shape:
        array.append(full_array[i, :])



from __future__ import annotations

from typing import Any, Callable, Iterable, Mapping, Optional

import numpy as np
import numpy.typing as npt

from .dataStructure import Aggregator


class NumpyProxyArray(np.ndarray):
    """A numpy array that dynamically updates attributes of container elements."""

    def __new__(
        cls, aggregator: SizedAggregator, instance: Any
    ) -> Optional[NumpyProxyArray]:
        """
        Create a new NumpyProxyArray instance using data from an aggregator.

        Parameters
        ----------
        aggregator : SizedAggregator
            The aggregator from which to obtain values.
        instance : Any
            The instance associated with the values.

        Returns
        -------
        Optional[NumpyProxyArray]
            A new instance of the class if values are available, otherwise None.
        """
        values = aggregator._get_values_from_container(instance, transpose=True)

        if values is None:
            return

        obj = values.view(cls)

        obj.aggregator = aggregator
        obj.instance = instance

        return obj

    def _get_values_from_aggregator(self) -> Any:
        """Refresh data from the underlying container."""
        return self.aggregator._get_values_from_container(self.instance, transpose=True, check=True)

    def __getitem__(self, index: int) -> Any:
        """Retrieve an item from the aggregated parameter array."""
        return self._get_values_from_aggregator()[index]

    def __setitem__(self, index: int, value: Any) -> None:
        """
        Modify an individual element in the aggregated parameter list.

        This ensures changes are propagated back to the objects.
        """
        current_value = self._get_values_from_aggregator()
        current_value[index] = value
        self.aggregator.__set__(self.instance, current_value)

    def __array_finalize__(self, obj: NumpyProxyArray) -> Optional[np.ndarray]:
        """Ensure attributes are copied when creating a new view or slice."""
        if obj is None:
            self.aggregator = None
            self.instance = None
            return

        if not isinstance(obj, NumpyProxyArray):
            return np.asarray(obj)

        self.aggregator = getattr(obj, "aggregator", None)
        self.instance = getattr(obj, "instance", None)

    def __array_function__(
        self,
        func: Callable,
        types: Iterable[type],
        *args: Iterable[Any],
        **kwargs: Mapping[str, Any],
    ) -> np.ndarray:
        """Ensure that high-level NumPy functions return a normal np.ndarray."""
        result = super().__array_function__(func, types, *args, **kwargs)
        return np.asarray(result)

    def __repr__(self) -> str:
        """Return a fresh representation that reflects live data."""
        return f"NumpyProxyA{self._get_values_from_aggregator().__repr__()[1:]}"


class SizedAggregator(Aggregator):
    """Aggregator for sized parameters."""

    def __init__(self, *args: Any, transpose: bool = False, **kwargs: Any) -> None:
        """
        Initialize a SizedAggregator instance.

        Parameters
        ----------
        *args : Any
            Variable length argument list.
        transpose : bool, options
            If False, the parameter shape will be ((n_instances, ) + parameter_shape).
            Else, it will be (parameter_shape + (n_instances, ))
            The default is False.
        **kwargs : Any
            Arbitrary keyword arguments.
        """
        self.transpose = transpose

        super().__init__(*args, **kwargs)

    def _parameter_shape(self, instance: Any) -> tuple[int, ...]:
        values = self._get_values_from_container(instance, transpose=False)

        shapes = [el.shape for el in values]

        if len(set(shapes)) > 1:
            raise ValueError("Inconsistent parameter shapes.")

        if len(shapes) == 0:
            return ()

        return shapes[0]

    def _expected_shape(self, instance: Any) -> tuple[int, ...]:
        if self.transpose:
            return self._parameter_shape(instance) + (self._n_instances(instance),)
        else:
            return (self._n_instances(instance),) + self._parameter_shape(instance)

    def _get_values_from_container(
        self, instance: Any, transpose: bool = False, check: bool = False
    ) -> np.ndarray:
        value = super()._get_values_from_container(instance, check=False)

        if value is None or len(value) == 0:
            return

        value = np.array(value, ndmin=2)

        if check:
            value = self._prepare(instance, value, transpose=False, recursive=True)
            self._check(instance, value, transpose=True, recursive=True)

        if transpose and self.transpose:
            value = value.T

        return value

    def _check(
        self,
        instance: Any,
        value: npt.ArrayLike,
        transpose: bool = True,
        recursive: bool = False,
    ) -> None:
        value_array = np.array(value, ndmin=2)
        if transpose and self.transpose:
            value_array = value_array.T

        value_shape = value_array.shape
        expected_shape = self._expected_shape(instance)

        if value_shape != expected_shape:
            raise ValueError(f"Expected a array with shape {expected_shape}, got {value_shape}")

        if recursive:
            super()._check(instance, value, recursive)

    def _prepare(
        self,
        instance: Any,
        value: npt.ArrayLike,
        transpose: bool = False,
        recursive: bool = False,
    ) -> np.ndarray:
        value = np.array(value, ndmin=2)

        if transpose and self.transpose:
            value = value.T

        if recursive:
            value = super()._prepare(instance, value, recursive)

        return value

    def __get__(self, instance: Any, cls: type) -> NumpyProxyArray:
        """
        Retrieve the descriptor value for the given instance.

        Parameters
        ----------
        instance : Any
            Instance to retrieve the descriptor value for.
        cls : Type[Any], optional
            Class to which the descriptor belongs.

        Returns
        -------
        NumpyProxyArray
            A numpy-like array that modifies the underlying objects when changed.
        """
        if instance is None:
            return self

        return NumpyProxyArray(self, instance)

    def __set__(self, instance: Any, value: Any) -> None:
        """
        Set the descriptor value for the given instance.

        Parameters
        ----------
        instance : Any
            Instance to set the descriptor value for.
        value : Any
            Value to set.
        """
        value = self._prepare(instance, value, transpose=True)
        super().__set__(instance, value)


class ClassDependentAggregator(Aggregator):
    """Aggregator where parameter name changes depending on instance type."""

    def __init__(self, *args: Any, mapping: dict, **kwargs: Any) -> None:
        """
        Initialize the Aggregator descriptor.

        Parameters
        ----------
        mapping : dict
            Mapping of instance types and parameter names.
        *args : tuple, optional
            Additional positional arguments.
        **kwargs : dict, optional
            Additional keyword arguments.
        """
        self.mapping = mapping

        super().__init__(*args, **kwargs)

    def _get_values_from_container(self, instance: Any, check: bool = False) -> list:
        container = self._container_obj(instance)

        values = []
        for el in container:
            if type(el) in self.mapping:
                attr = self.mapping[type(el)]
            else:
                attr = self.mapping[None]
                if attr is None:
                    continue

            value = getattr(el, attr)
            values.append(value)

        if len(values) == 0:
            return

        if check:
            values = self._prepare(instance, values, transpose=False, recursive=True)
            self._check(instance, values, transpose=True, recursive=True)

        return values


class SizedClassDependentAggregator(SizedAggregator, ClassDependentAggregator):
    """Aggregator where parameter name and size changes depending on instance type."""

    pass



# ---
# jupyter:
#   jupytext:
#     formats: md:myst,py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.17.1
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Model System Periphery
#
# Show Foto
#
# Show Abstraction
#
# Show Model
#
#
# Fit Parameters
#
# Missing:
# - Data
# - Fotos



from __future__ import annotations

import warnings
from functools import wraps
from typing import Any, Optional

import numpy as np

from CADETProcess.dataStructure import (
    Aggregator,
    Bool,
    SizedAggregator,
    SizedClassDependentAggregator,
    SizedNdArray,
    String,
    Structure,
    UnsignedFloat,
    deprecated_alias,
)

from .componentSystem import ComponentSystem


class Reaction(Structure):
    """
    Helper class to store information about individual Mass Action Law Reactions.

    This class represents an individual reaction within a MAL model, and
    stores information about the reaction's stoichiometry, rate constants,
    and concentration exponents. It is used to create a ReactionScheme object,
    which contains information about all of the reactions in the MAL model.

    Attributes
    ----------
    component_system : ComponentSystem
        Component system of the reaction.
    k_fwd : float
        Forward reaction rate.
    k_bwd : float
        Backward reaction rate.
    is_kinetic : bool
        Flag indicating whether the reaction is kinetic (i.e., whether the
        reaction rates are explicitly defined) or whether the reaction is
        assumed to be at rapid equilibrium.
    k_fwd_min : float
        Minimum value of the forward reaction rate in case of rapid equilibrium.
    exponents_fwd : list of float
        Concentration exponents of the components in order of indices for
        forward reaction.
    exponents_bwd : list of float
        Concentration exponents of the components in order of indices for
        backward reaction.
    stoich : np.ndarray
        Stoichiometric coefficients of the components in the reaction.
    n_comp : int
        The number of components in the reaction.
    k_eq : float
        The equilibrium constant for the reaction.
    """

    is_kinetic = Bool(default=True)
    stoich = SizedNdArray(size="n_comp")
    k_fwd = UnsignedFloat()
    k_bwd = UnsignedFloat()
    k_fwd_min = UnsignedFloat(default=100)
    exponents_fwd = SizedNdArray(size="n_comp", default=0)
    exponents_bwd = SizedNdArray(size="n_comp", default=0)

    _parameters = [
        "is_kinetic",
        "stoich",
        "k_fwd",
        "k_bwd",
        "k_fwd_min",
        "exponents_fwd",
        "exponents_bwd",
    ]

    @deprecated_alias(indices="components")
    def __init__(
        self,
        component_system: ComponentSystem,
        components: list[int | str],
        coefficients: np.ndarray,
        k_fwd: float,
        k_bwd: float = 1,
        is_kinetic: bool = True,
        k_fwd_min: float = 100,
        exponents_fwd: Optional[list[float]] = None,
        exponents_bwd: Optional[list[float]] = None,
    ) -> None:
        """
        Initialize individual Mass Action Law Reaction.

        Parameters
        ----------
        component_system : ComponentSystem
            Component system of the reaction.
        components : list of int or strings
            Component names of the components involved in the reaction.
        coefficients : np.ndarray
            Stoichiometric coefficients in the same order of components .
        k_fwd : float
            Forward reaction rate.
        k_bwd : float, optional
            Backward reaction rate. The default is 1.
        is_kinetic : bool, optional
            If False, reaction rates are scaled up to approximate rapid
            equilibriums. The default is True.
        k_fwd_min : float, optional
            Minimum value of foward reaction rate in case of rapid equilbrium.
            The default is 100.
        exponents_fwd : list of float, optional
            Concentration exponents of the components in order of components for
            forward reaction. If None is given, values are inferred from the
            stoichiometric coefficients. The default is None.
        exponents_bwd : list of float, optional
            Concentration exponents of the components in order of components for
            backward reaction. If None is given, values are inferred from the
            stoichiometric coefficients. The default is None.
        """
        self.component_system = component_system
        super().__init__()

        self.is_kinetic = is_kinetic
        if not is_kinetic:
            k_fwd, k_bwd = scale_to_rapid_equilibrium(k_fwd, k_fwd_min)

        self.k_fwd = k_fwd
        self.k_bwd = k_bwd

        if isinstance(components[0], str):
            indices = [component_system.species.index(i) for i in components]
        else:
            warnings.warn(
                "Component are expected to be specified by name. "
                "This will be deprecated in future versions.",
                DeprecationWarning,
                stacklevel=2,
            )
            indices = components

        self.stoich = np.zeros((self.n_comp,))
        for i, c in zip(indices, coefficients):
            self.stoich[i] = c

        if exponents_fwd is None:
            e_fwd = np.maximum(np.zeros((self.n_comp,)), -self.stoich)
        else:
            e_fwd = np.zeros((self.n_comp,))
            for i, e in zip(indices, exponents_fwd):
                e_fwd[i] = e
        self.exponents_fwd = e_fwd

        if exponents_bwd is None:
            e_bwd = np.maximum(np.zeros((self.n_comp,)), self.stoich)
        else:
            e_bwd = np.zeros((self.n_comp,))
            for i, e in zip(indices, exponents_bwd):
                e_bwd[i] = e
        self.exponents_bwd = e_bwd

    @property
    def n_comp(self) -> int:
        """int: Number of components."""
        return self.component_system.n_comp

    @property
    def k_eq(self) -> float:
        """float: Equilibrium constant (Ratio of forward and backward reaction)."""
        return self.k_fwd / self.k_bwd

    def __str__(self) -> str:
        """str: String representation of the Reaction."""
        educts = []
        products = []
        for i, nu in enumerate(self.stoich):
            if nu < 0:
                if nu == -1:
                    educts.append(f"{self.component_system.species[i]}")
                else:
                    educts.append(f"{abs(nu)} {self.component_system.species[i]}")
            elif nu > 0:
                if nu == 1:
                    products.append(f"{self.component_system.species[i]}")
                else:
                    products.append(f"{nu} {self.component_system.species[i]}")

        if self.is_kinetic:
            reaction_operator = f" <=>[{self.k_fwd:.2E}][{self.k_bwd:.2E}] "
        else:
            reaction_operator = f" <=>[{self.k_eq:.2E}] "

        return " + ".join(educts) + reaction_operator + " + ".join(products)


class CrossPhaseReaction(Structure):
    """
    Helper class to store information about cross-phase Mass Action Law reactions.

    Attributes
    ----------
    component_system : ComponentSystem
        The component system of the reaction.
    stoich_liquid : ndarray
        An array of stoichiometric coefficients of the components in the liquid phase.
    stoich_solid : ndarray
        An array of stoichiometric coefficients of the components in the solid phase.
    exponents_fwd_solid_modliquid : ndarray
        An array of concentration exponents of the components in the solid phase for the
        forward reaction in the liquid phase.
    exponents_bwd_solid_modliquid : ndarray
        An array of concentration exponents of the components in the solid phase for the
        backward reaction in the liquid phase.
    exponents_fwd_liquid_modsolid : ndarray
        An array of concentration exponents of the components in the liquid phase for
        the forward reaction in the solid phase.
    exponents_bwd_liquid_modsolid : ndarray
        An array of concentration exponents of the components in the liquid phase for
        the backward reaction in the solid phase.
    is_kinetic : bool
        A boolean flag indicating whether the reaction is kinetic or not.
    k_fwd : float
        The forward reaction rate.
    k_bwd : float
        The backward reaction rate.
    exponents_fwd_liquid : ndarray
        An array of concentration exponents of the components in the liquid phase for
        the forward reaction.
    exponents_bwd_liquid : ndarray
        An array of concentration exponents of the components in the liquid phase for
        the backward reaction.
    exponents_fwd_solid : ndarray
        An array of concentration exponents of the components in the solid phase for the
        forward reaction.
    exponents_bwd_solid : ndarray
        An array of concentration exponents of the components in the solid phase for the
        backward reaction.
    """

    is_kinetic = Bool(default=True)

    stoich_liquid = SizedNdArray(size="n_comp", default=0)
    stoich_solid = SizedNdArray(size="n_comp", default=0)
    k_fwd = UnsignedFloat()
    k_bwd = UnsignedFloat()
    k_fwd_min = UnsignedFloat(default=100)

    exponents_fwd_liquid = SizedNdArray(size="n_comp", default=0)
    exponents_fwd_solid = SizedNdArray(size="n_comp", default=0)

    exponents_bwd_liquid = SizedNdArray(size="n_comp", default=0)
    exponents_bwd_solid = SizedNdArray(size="n_comp", default=0)

    exponents_fwd_liquid_modsolid = SizedNdArray(size="n_comp", default=0)
    exponents_fwd_solid_modliquid = SizedNdArray(size="n_comp", default=0)

    exponents_bwd_liquid_modsolid = SizedNdArray(size="n_comp", default=0)
    exponents_bwd_solid_modliquid = SizedNdArray(size="n_comp", default=0)

    _parameters = [
        "stoich_liquid",
        "stoich_solid",
        "k_fwd",
        "k_bwd",
        "k_fwd_min",
        "exponents_fwd_liquid",
        "exponents_fwd_solid",
        "exponents_bwd_liquid",
        "exponents_bwd_solid",
        "exponents_fwd_liquid_modsolid",
        "exponents_fwd_solid_modliquid",
        "exponents_bwd_liquid_modsolid",
        "exponents_bwd_solid_modliquid",
    ]

    @deprecated_alias(indices="components")
    def __init__(
        self,
        component_system: ComponentSystem,
        components: list[list | str],
        coefficients: list[float],
        phases: list[int],
        k_fwd: float,
        k_bwd: float = 1,
        is_kinetic: bool = True,
        k_fwd_min: float = 100,
        exponents_fwd_liquid: Optional[list[float]] = None,
        exponents_bwd_liquid: Optional[list[float]] = None,
        exponents_fwd_solid: Optional[list[float]] = None,
        exponents_bwd_solid: Optional[list[float]] = None,
    ) -> None:
        """
        Initialize individual cross-phase MAL reaction.

        Parameters
        ----------
        component_system : ComponentSystem
            Component system of the reaction.
        components : list of int or strings
            Component names of the components involved in the reaction.
        coefficients : list
            Stoichiometric coefficients in the same order of component indices.
        phases : list
            phase indices of the component.
            0: liquid phase
            1: solid phase
        k_fwd : float
            Forward reaction rate.
        k_bwd : float, optional
            Backward reaction rate. The default is 1.
        is_kinetic : Bool, optional
            If False, reaction rates are scaled up to approximate rapid
            equilibriums. The default is True.
        k_fwd_min : float, optional
            Minimum value of foward reaction rate in case of rapid equilbrium.
            The default is 100.
        exponents_fwd_liquid : list, optional
            Concentration exponents of the components in order of indices for
            forward reaction in liquid phase. If None is given, values are
            inferred from the stoichiometric coefficients. The default is None.
        exponents_bwd_liquid : list, optional
            Concentration exponents of the components in order of indices for
            backward reaction in liquid phase. If None is given, values are
            inferred from the stoichiometric coefficients. The default is None.
        exponents_fwd_solid : list, optional
            Concentration exponents of the components in order of indices for
            forward reaction in solid phase. If None is given, values are
            inferred from the stoichiometric coefficients. The default is None.
        exponents_bwd_solid : list, optional
            Concentration exponents of the components in order of indices for
            backward reaction in solid phase. If None is given, values are
            inferred from the stoichiometric coefficients. The default is None.
        """
        self.component_system = component_system
        super().__init__()

        self.stoich_liquid = np.zeros((self.n_comp,))
        self.stoich_solid = np.zeros((self.n_comp,))
        self.exponents_fwd_solid_modliquid = np.zeros((self.n_comp,))
        self.exponents_bwd_solid_modliquid = np.zeros((self.n_comp,))
        self.exponents_fwd_liquid_modsolid = np.zeros((self.n_comp,))
        self.exponents_bwd_liquid_modsolid = np.zeros((self.n_comp,))

        if isinstance(components[0], str):
            indices = [component_system.species.index(i) for i in components]
        else:
            warnings.warn(
                "Component are expected to be specified by name. "
                "This will be deprecated in future versions.",
                DeprecationWarning,
                stacklevel=2,
            )
            indices = components

        if phases is None:
            phases = [0 for n in indices]

        for i, c, p in zip(indices, coefficients, phases):
            if p == 0:
                self.stoich_liquid[i] = c
                if c < 0:
                    self.exponents_fwd_solid_modliquid[i] = abs(c)
                elif c > 0:
                    self.exponents_bwd_solid_modliquid[i] = c
            elif p == 1:
                self.stoich_solid[i] = c
                if c < 0:
                    self.exponents_fwd_liquid_modsolid[i] = abs(c)
                elif c > 0:
                    self.exponents_bwd_liquid_modsolid[i] = c

        self.is_kinetic = is_kinetic
        if not is_kinetic:
            k_fwd, k_bwd = scale_to_rapid_equilibrium(k_fwd, k_fwd_min)

        self.k_fwd = k_fwd
        self.k_bwd = k_bwd

        if exponents_fwd_liquid is None:
            e_fwd = np.maximum(np.zeros((self.n_comp,)), -self.stoich_liquid)
        else:
            e_fwd = np.zeros((self.n_comp,))
            for i, e in zip(indices, exponents_fwd_liquid):
                e_fwd[i] = e
        self.exponents_fwd_liquid = e_fwd

        if exponents_bwd_liquid is None:
            e_bwd = np.maximum(np.zeros((self.n_comp,)), self.stoich_liquid)
        else:
            e_bwd = np.zeros((self.n_comp,))
            for i, e in zip(indices, exponents_bwd_liquid):
                e_bwd[i] = e
        self.exponents_bwd_liquid = e_bwd

        if exponents_fwd_solid is None:
            e_fwd = np.maximum(np.zeros((self.n_comp,)), -self.stoich_solid)
        else:
            e_fwd = np.zeros((self.n_comp,))
            for i, e in zip(indices, exponents_fwd_solid):
                e_fwd[i] = e
        self.exponents_fwd_solid = e_fwd

        if exponents_bwd_solid is None:
            e_bwd = np.maximum(np.zeros((self.n_comp,)), self.stoich_solid)
        else:
            e_bwd = np.zeros((self.n_comp,))
            for i, e in zip(indices, exponents_bwd_solid):
                e_bwd[i] = e
        self.exponents_bwd_solid = e_bwd

    @property
    def n_comp(self) -> int:
        """int: Number of components."""
        return self.component_system.n_comp

    @property
    def k_eq(self) -> float:
        """float: Equilibrium constant (Ratio of forward and backward reaction)."""
        return self.k_fwd / self.k_bwd

    def __str__(self) -> str:
        """str: String representation of the Reaction."""
        educts = []
        products = []
        for i, nu in enumerate(self.stoich_liquid):
            if nu < 0:
                if nu == -1:
                    educts.append(f"{self.component_system.species[i]}(l)")
                else:
                    educts.append(f"{abs(nu)} {self.component_system.species[i]}(l)")
            elif nu > 0:
                if nu == 1:
                    products.append(f"{self.component_system.species[i]}(l)")
                else:
                    products.append(f"{nu} {self.component_system.species[i]}(l)")
        for i, nu in enumerate(self.stoich_solid):
            if nu < 0:
                if nu == -1:
                    educts.append(f"{self.component_system.species[i]}(s)")
                else:
                    educts.append(f"{abs(nu)} {self.component_system.species[i]}(s)")
            elif nu > 0:
                if nu == 1:
                    products.append(f"{self.component_system.species[i]}(s)")
                else:
                    products.append(f"{nu} {self.component_system.species[i]}(s)")

        if self.is_kinetic:
            reaction_operator = f" <=>[{self.k_fwd:.2E}][{self.k_bwd:.2E}] "
        else:
            reaction_operator = f" <=>[{self.k_eq:.2E}] "

        return " + ".join(educts) + reaction_operator + " + ".join(products)


class ReactionBaseClass(Structure):
    """
    Abstract base class for parameters of reaction models.

    Attributes
    ----------
    parameters : dict
        dict with parameter values.
    name : String
        name of the reaction model.
    """

    name = String()

    _parameters = []

    def __init__(
        self,
        component_system: ComponentSystem,
        name: Optional[str] = None,
        *args: Any,
        **kwargs: Any,
    ) -> None:
        """Initialize reaction base."""
        self.component_system = component_system
        self.name = name

        super().__init__(*args, **kwargs)

    @property
    def model(self) -> str:
        """str: Name of the reaction model."""
        return self.__class__.__name__

    @property
    def component_system(self) -> ComponentSystem:
        """ComponentSystem: Component System."""
        return self._component_system

    @component_system.setter
    def component_system(self, component_system: ComponentSystem) -> None:
        if not isinstance(component_system, ComponentSystem):
            raise TypeError("Expected ComponentSystem")
        self._component_system = component_system

    @property
    def n_comp(self) -> int:
        """int: Number of components."""
        return self.component_system.n_comp

    def __repr__(self) -> str:
        """str: String representation of the Reaction."""
        return f"{self.__class__.__name__}(n_comp={self.n_comp}, name={self.name})"

    def __str__(self) -> str:
        """str: Name of the Reaction."""
        if self.name is None:
            return self.__class__.__name__
        return self.name


class NoReaction(ReactionBaseClass):
    """
    Dummy class for units that do not experience reaction behavior.

    The number of components is set to zero for this class.
    """

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize NoReaction."""
        super().__init__(ComponentSystem(), name="NoReaction")


class BulkReactionBase(ReactionBaseClass):
    """Base class for bulk reaction systems."""

    @classmethod
    def to_particle_model() -> ParticleReactionBase:
        """Convert bulk reaction model to particle reaction model."""
        raise NotImplementedError


class MassActionLaw(BulkReactionBase):
    """Parameters for Reaction in Bulk Phase."""

    k_fwd = Aggregator("k_fwd", "reactions")
    k_bwd = Aggregator("k_bwd", "reactions")
    stoich = SizedAggregator("stoich", "reactions", transpose=True)
    exponents_fwd = SizedAggregator("exponents_fwd", "reactions", transpose=True)
    exponents_bwd = SizedAggregator("exponents_bwd", "reactions", transpose=True)

    _parameters = ["stoich", "exponents_fwd", "exponents_bwd", "k_fwd", "k_bwd"]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize MassActionLaw."""
        self._reactions = []
        super().__init__(*args, **kwargs)

    @wraps(Reaction.__init__)
    def add_reaction(self, *args: Any, **kwargs: Any) -> Reaction:
        """Add reaction to ReactionSystem."""
        r = Reaction(self.component_system, *args, **kwargs)
        self._reactions.append(r)

        return r

    @property
    def reactions(self) -> int:
        """list: Reactions in ReactionSystem."""
        return self._reactions

    @property
    def n_reactions(self) -> int:
        """int: Number of Reactions."""
        return len(self.reactions)

    @property
    def k_eq(self) -> list:
        """list: Equilibrium constants of liquid phase Reactions."""
        return [r.k_eq for r in self.reactions]

    def to_particle_model(self) -> "MassActionLawParticle":
        """Convert Bulk Reaction Model to Particle Reaction Model."""
        particle_model = MassActionLawParticle(self.component_system, self.name)
        particle_model._liquid_reactions = self.reactions

        return particle_model


class ParticleReactionBase(ReactionBaseClass):
    """Base class for bulk reaction systems."""


class MassActionLawParticle(ParticleReactionBase):
    """Parameters for Reaction in Particle Phase."""

    stoich_liquid = SizedClassDependentAggregator(
        "stoich_liquid",
        "liquid_reactions",
        mapping={CrossPhaseReaction: "stoich_liquid", None: "stoich"},
        transpose=True,
    )
    k_fwd_liquid = Aggregator("k_fwd", "liquid_reactions")
    k_bwd_liquid = Aggregator("k_bwd", "liquid_reactions")
    exponents_fwd_liquid = SizedAggregator(
        "exponents_fwd", "liquid_reactions", transpose=True
    )
    exponents_bwd_liquid = SizedAggregator(
        "exponents_bwd", "liquid_reactions", transpose=True
    )

    stoich_solid = SizedClassDependentAggregator(
        "stoich_solid",
        "solid_reactions",
        mapping={CrossPhaseReaction: "stoich_solid", None: "stoich"},
        transpose=True,
    )
    k_fwd_solid = Aggregator("k_fwd", "solid_reactions")
    k_bwd_solid = Aggregator("k_bwd", "solid_reactions")
    exponents_fwd_solid = SizedAggregator(
        "exponents_fwd", "solid_reactions", transpose=True
    )
    exponents_bwd_solid = SizedAggregator(
        "exponents_bwd", "solid_reactions", transpose=True
    )

    exponents_fwd_liquid_modsolid = SizedClassDependentAggregator(
        "exponents_fwd_liquid_modsolid",
        "liquid_reactions",
        mapping={CrossPhaseReaction: "exponents_fwd_liquid_modsolid", None: None},
        transpose=True,
    )
    exponents_bwd_liquid_modsolid = SizedClassDependentAggregator(
        "exponents_bwd_liquid_modsolid",
        "liquid_reactions",
        mapping={CrossPhaseReaction: "exponents_bwd_liquid_modsolid", None: None},
        transpose=True,
    )

    exponents_fwd_solid_modliquid = SizedClassDependentAggregator(
        "exponents_fwd_solid_modliquid",
        "solid_reactions",
        mapping={CrossPhaseReaction: "exponents_fwd_solid_modliquid", None: None},
        transpose=True,
    )
    exponents_bwd_solid_modliquid = SizedClassDependentAggregator(
        "exponents_bwd_solid_modliquid",
        "solid_reactions",
        mapping={CrossPhaseReaction: "exponents_bwd_solid_modliquid", None: None},
        transpose=True,
    )

    _parameters = [
        "stoich_liquid",
        "exponents_fwd_liquid",
        "exponents_bwd_liquid",
        "k_fwd_liquid",
        "k_bwd_liquid",
        "exponents_fwd_liquid_modsolid",
        "exponents_bwd_liquid_modsolid",
        "stoich_solid",
        "exponents_fwd_solid",
        "exponents_bwd_solid",
        "k_fwd_solid",
        "k_bwd_solid",
        "exponents_fwd_solid_modliquid",
        "exponents_bwd_solid_modliquid",
    ]

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        """Initialize MassActionLawParticle."""
        self._liquid_reactions = []
        self._solid_reactions = []
        self._cross_phase_reactions = []

        super().__init__(*args, **kwargs)

    @wraps(Reaction.__init__)
    def add_liquid_reaction(self, *args: Any, **kwargs: Any) -> None:
        """Add liquid phase to ReactionSystem."""
        r = Reaction(self.component_system, *args, **kwargs)
        self._liquid_reactions.append(r)

    @wraps(Reaction.__init__)
    def add_solid_reaction(self, *args: Any, **kwargs: Any) -> None:
        """Add solid phase to ReactionSystem."""
        r = Reaction(self.component_system, *args, **kwargs)
        self._solid_reactions.append(r)

    @wraps(CrossPhaseReaction.__init__)
    def add_cross_phase_reaction(self, *args: Any, **kwargs: Any) -> None:
        """Add cross phase to ReactionSystem."""
        r = CrossPhaseReaction(self.component_system, *args, **kwargs)
        self._cross_phase_reactions.append(r)

    # Pore Liquid
    @property
    def liquid_reactions(self) -> list:
        """list: Liquid phase Reactions."""
        return self._liquid_reactions + self.cross_phase_reactions

    @property
    def n_liquid_reactions(self) -> int:
        """int: Number of liquid phase Reactions."""
        return len(self.liquid_reactions)

    @property
    def k_eq_liquid(self) -> list:
        """list: Equilibrium constants of liquid phase Reactions."""
        return [r.k_eq for r in self.liquid_reactions]

    # Solid
    @property
    def solid_reactions(self) -> list:
        """list: Solid phase Reactions."""
        return self._solid_reactions + self.cross_phase_reactions

    @property
    def n_solid_reactions(self) -> int:
        """int: Number of solid phase Reactions."""
        return len(self.solid_reactions)

    @property
    def k_eq_solid(self) -> list:
        """list: Equilibrium constants of solid phase Reactions."""
        return [r.k_eq for r in self.solid_reactions]

    # Cross Phase
    @property
    def cross_phase_reactions(self) -> list:
        """list: Cross phase reactions."""
        return self._cross_phase_reactions

    @property
    def n_cross_phase_reactions(self) -> int:
        """int: Number of cross phase Reactions."""
        return len(self.cross_phase_reactions)


def scale_to_rapid_equilibrium(
    k_eq: float,
    k_fwd_min: float = 10,
) -> tuple[float, float]:
    """
    Scale forward and backward reaction rates if only k_eq is known.

    Parameters
    ----------
    k_eq : float
       Equilibrium constant.
    k_fwd_min : float, optional
        Minimum value for forwards reaction. The default is 10.

    Returns
    -------
    k_fwd : float
        Forward reaction rate.
    k_bwd : float
        Backward reaction rate.
    """
    k_fwd = k_fwd_min
    k_bwd = k_fwd_min / k_eq

    return k_fwd, k_bwd



"""
=============================================
Performance (:mod:`CADETProcess.performance`)
=============================================

.. currentmodule:: CADETProcess.performance


Performance data
================

Classes for storing the performance parameters after fractionation.

.. autosummary::
    :toctree: generated/

    Performance
    RankedPerformance

Performance Indicators
======================

Individual performance indicators (extracted from Performance).
Mostly convenince method.

.. autosummary::
    :toctree: generated/

    PerformanceIndicator
    Mass
    Recovery
    Productivity
    EluentConsumption
    Purity
    Concentration
    PerformanceProduct
    MassBalanceDifference

Notes
-----
Performance Indicators might be deprecated in future since with new evaluation chains
it's no longer required for setting up as optimization problem.

"""  # noqa

from typing import Any, Optional

import numpy as np

from CADETProcess import CADETProcessError
from CADETProcess.dataStructure import SizedNdArray, Structure
from CADETProcess.metric import MetricBase
from CADETProcess.processModel import ComponentSystem


class Performance(Structure):
    """
    Class for storing the performance parameters after fractionation.

    Attributes
    ----------
    mass : np.ndarray
        Mass of each component in the system after fractionation.
        Size depends on number of components.
    concentration : np.ndarray
        Concentration of each component in the system after fractionation.
        Size depends on number of components.
    purity : np.ndarray
        Purity of each component in the system after fractionation.
        Size depends on number of components.
    recovery : np.ndarray
        Recovery of each component in the system after fractionation.
        Size depends on number of components.
    productivity : np.ndarray
        Productivity of each component in the system after fractionation.
        Size depends on number of components.
    eluent_consumption : np.ndarray
        Eluent consumption of each component in the system after fractionation.
        Size depends on number of components.
    mass_balance_difference : np.ndarray
        Mass balance difference of each component.
        Size depends on number of components.
    component_system : ComponentSystem
        The component system used for fractionation.
        If not provided, a default component system is used.

    See Also
    --------
    CADETProcess.fractionation
    RankedPerformance
    """

    _performance_keys = [
        "mass",
        "concentration",
        "purity",
        "recovery",
        "productivity",
        "eluent_consumption",
        "mass_balance_difference",
    ]

    mass = SizedNdArray(size=("n_comp"))
    concentration = SizedNdArray(size=("n_comp"))
    purity = SizedNdArray(size=("n_comp"))
    recovery = SizedNdArray(size=("n_comp"))
    productivity = SizedNdArray(size=("n_comp"))
    eluent_consumption = SizedNdArray(size=("n_comp"))
    mass_balance_difference = SizedNdArray(size=("n_comp"))

    def __init__(
        self,
        mass: np.ndarray,
        concentration: np.ndarray,
        purity: np.ndarray,
        recovery: np.ndarray,
        productivity: np.ndarray,
        eluent_consumption: np.ndarray,
        mass_balance_difference: np.ndarray,
        component_system: Optional[ComponentSystem] = None,
    ) -> None:
        """
        Initialize Performance.

        Parameters
        ----------
        mass : ndarray
            The mass of each component.
        concentration : ndarray
            The concentration of each component.
        purity : ndarray
            The purity of each component.
        recovery : ndarray
            The recovery of each component.
        productivity : ndarray
            The productivity of each component.
        eluent_consumption : ndarray
            The eluent consumption of each component.
        mass_balance_difference : ndarray
            The difference in mass balance of each component.
        component_system : ComponentSystem
            The ComponentSystem object that describes the system's components.
        """
        if component_system is None:
            component_system = ComponentSystem(mass.shape[0])

        self.component_system = component_system
        self.mass = mass
        self.concentration = concentration
        self.purity = purity
        self.recovery = recovery
        self.productivity = productivity
        self.eluent_consumption = eluent_consumption
        self.mass_balance_difference = mass_balance_difference

    @property
    def n_comp(self) -> int:
        """int: Number of components in the system."""
        return self.component_system.n_comp

    def to_dict(self) -> dict:
        """Return dictionary representation of the object."""
        return {key: getattr(self, key).tolist() for key in self._performance_keys}

    def __getitem__(self, item: str) -> Any:
        """Get an attribute of the object by its name."""
        if item not in self._performance_keys:
            raise AttributeError("Not a valid performance parameter")

        return getattr(self, item)

    def __repr__(self) -> str:
        """str: String representation of the object."""
        return (
            f"{self.__class__.__name__}(mass={np.array_repr(self.mass)}, "
            f"concentration={np.array_repr(self.concentration)}, "
            f"purity={np.array_repr(self.purity)}, "
            f"recovery={np.array_repr(self.recovery)}, "
            f"productivity={np.array_repr(self.productivity)}, "
            f"eluent_consumption={np.array_repr(self.eluent_consumption)} "
            f"mass_balance_difference={np.array_repr(self.mass_balance_difference)})"
        )


class RankedPerformance:
    """
    Class for calculating a weighted average of the Performance.

    See Also
    --------
    Performance
    ranked_objective_decorator
    """

    _performance_keys = Performance._performance_keys

    def __init__(
        self,
        performance: Performance,
        ranking: Optional[float] = 1.0,
    ) -> None:
        """Initialize RankedPerformance."""
        if not isinstance(performance, Performance):
            raise TypeError("Expected Performance")

        self._performance = performance

        self.ranking = ranking

    @property
    def performance(self) -> Performance:
        """Performance: Performance object."""
        return self._performance

    @property
    def ranking(self) -> float:
        """list[float]: Relative weighting factors for multi component evaluation."""
        return self._ranking

    @ranking.setter
    def ranking(self, ranking: tuple[float | int] | np.ndarray) -> None:
        if isinstance(ranking, (float, int)):
            ranking = self.performance.n_comp * [ranking]
        elif len(ranking) != self.performance.n_comp:
            raise CADETProcessError("Number of components does not match.")

        self._ranking = ranking

    def to_dict(self) -> dict:
        """Return dictionary representation of the object."""
        return {key: float(getattr(self, key)) for key in self._performance_keys}

    def __getattr__(self, item: str) -> float:
        """Retrieve a performance attribute by its name, weighted by ranking."""
        if item not in self._performance_keys:
            raise AttributeError
        return sum(self._performance[item] * self.ranking) / sum(self.ranking)

    def __getitem__(self, item: str) -> Any:
        """Retrieve an attribute of the object by its name using indexing syntax."""
        if item not in self._performance_keys:
            raise AttributeError("Not a valid performance parameter")
        return getattr(self, item)

    def __repr__(self) -> str:
        """str: Sting representation of the object."""
        return (
            f"{self.__class__.__name__}(mass={np.array_repr(self.mass)}, "
            f"concentration={np.array_repr(self.concentration)}, "
            f"purity={np.array_repr(self.purity)}, "
            f"recovery={np.array_repr(self.recovery)}, "
            f"productivity={np.array_repr(self.productivity)}, "
            f"eluent_consumption={np.array_repr(self.eluent_consumption)} "
            f"mass_balance_difference={np.array_repr(self.mass_balance_difference)})"
        )


class PerformanceIndicator(MetricBase):
    """
    Base class for performance indicators used in optimization and fractionation.

    See Also
    --------
    RankedPerformance
    """

    def __init__(self, ranking: Optional[list[float]] = None) -> None:
        """
        Initialize PerformanceIndicator.

        Parameters
        ----------
        ranking : list of floats, optional
            Weights to rank individual compoments. If None, all compoments are ranke
        """
        self.ranking = ranking

    @property
    def ranking(self) -> float:
        """list[float]: Relative weighting factors for multi component evaluation."""
        return self._ranking

    @ranking.setter
    def ranking(self, ranking: list[float]) -> None:
        self._ranking = ranking

    @property
    def bad_metrics(self) -> int:
        """int: Bad metrics to use when evaluation fails."""
        return 0

    def evaluate(self, performance: Performance) -> list:
        """
        Evaluate the performance indicator for the given performance data.

        Parameters
        ----------
        performance : Performance
            Object containing performance data.

        Returns
        -------
        list
            List of performance indicator values.
        """
        try:
            performance = performance.performance
        except AttributeError:
            pass

        if self.ranking is not None:
            performance = RankedPerformance(performance, self.ranking)

        value = self._evaluate(performance).tolist()

        if self.ranking is not None:
            metric = [value]
        else:
            metric = []
            for i, comp in enumerate(performance.component_system):
                metric.append(value[i])

        return metric

    __call__ = evaluate

    def __str__(self) -> str:
        """str: String representation of the class."""
        return self.__class__.__name__


class Mass(PerformanceIndicator):
    """
    Performance indicator based on the mass of each component in the system.

    See Also
    --------
    PerformanceIndicator
    """

    def _evaluate(self, performance: Performance) -> np.ndarray:
        return performance.mass


class Recovery(PerformanceIndicator):
    """
    Performance indicator based on the recovery of each component in the system.

    See Also
    --------
    PerformanceIndicator
    """

    def _evaluate(self, performance: Performance) -> np.ndarray:
        return performance.recovery


class Productivity(PerformanceIndicator):
    """
    Performance indicator based on the productivity of each component in the system.

    See Also
    --------
    PerformanceIndicator
    """

    def _evaluate(self, performance: Performance) -> np.ndarray:
        return performance.productivity


class EluentConsumption(PerformanceIndicator):
    """
    Performance indicator based on the specific eluent consumption of each component.

    See Also
    --------
    PerformanceIndicator
    """

    def _evaluate(self, performance: Performance) -> np.ndarray:
        return performance.eluent_consumption


class Purity(PerformanceIndicator):
    """
    Performance indicator based on the purity of each component in the system.

    See Also
    --------
    PerformanceIndicator
    """

    def _evaluate(self, performance: Performance) -> np.ndarray:
        return performance.purity


class Concentration(PerformanceIndicator):
    """
    Performance indicator based on the concentration of each component in the system.

    See Also
    --------
    PerformanceIndicator
    """

    def _evaluate(self, performance: Performance) -> np.ndarray:
        return performance.concentration


class PerformanceProduct(PerformanceIndicator):
    """
    Performance indicator based on the product of several performance indicators.

    See Also
    --------
    Productivity
    Recovery
    EluentConsumption
    PerformanceIndicator
    """

    def _evaluate(self, performance: Performance) -> np.ndarray:
        return (
            performance.productivity
            * performance.recovery
            * performance.eluent_consumption
        )


class MassBalanceDifference(PerformanceIndicator):
    """
    Performance indicator based on the mass balance of each component in the system.

    See Also
    --------
    PerformanceIndicator
    """

    def _evaluate(self, performance: Performance) -> np.ndarray:
        return np.abs(performance.mass_balance_difference)



import unittest

import numpy as np
from CADETProcess.optimization import Individual, hash_array


def setup_individual(n_vars=2, n_obj=1, n_nonlin=0, n_meta=0, rng=None):
    if rng is None:
        rng = np.random.default_rng(12345)

    x = rng.random(n_vars)
    f = -rng.random(n_obj)
    if n_nonlin > 0:
        g = np.random.random(n_nonlin)
    else:
        g = None

    if n_nonlin > 0:
        m = np.sum(f)
    else:
        m = None

    return Individual(x, f=f, g=g, m=m)


class TestHashArray(unittest.TestCase):
    def test_hash_array(self):
        array = np.array([1, 2.0])
        expected_hash = 'dc91ce9a50ddc828740aa26743716897fdb2bb64f1db662fe263a59be56145ae'
        self.assertEqual(hash_array(array), expected_hash)


class TestIndividual(unittest.TestCase):
    def setUp(self):
        x = [1, 2]
        f = [-1]
        self.individual_1 = Individual(x, f)

        x = [1, 2]
        f = [-1.001]
        self.individual_2 = Individual(x, f)

        x = [1, 2]
        f = [-1, -2]
        self.individual_multi_1 = Individual(x, f)

        x = [1, 2]
        f = [-1.001, -2]
        self.individual_multi_2 = Individual(x, f)

        x = [1, 2]
        f = [-1, -2.001]
        self.individual_multi_3 = Individual(x, f)

        x = [1, 2]
        f = [-1]
        g = [3]
        self.individual_constr_1 = Individual(x, f, g)

        x = [2, 3]
        f = [-1]
        g = [3]
        self.individual_constr_2 = Individual(x, f, g)

        x = [3, 4]
        f = [-1]
        g = [2]
        self.individual_constr_3 = Individual(x, f, g)

        x = [3, 4]
        f = [-1]
        g = [2]
        m = [1]
        self.individual_constr_meta = Individual(x, f, g, cv_nonlincon=g, m=m)

    def test_dimensions(self):
        dimensions_expected = (2, 1, 0, 0)
        dimensions = self.individual_1.dimensions
        self.assertEqual(dimensions, dimensions_expected)

        self.assertEqual(2, self.individual_1.n_x)
        self.assertEqual(1, self.individual_1.n_f)
        self.assertEqual(0, self.individual_1.n_g)
        self.assertEqual(0, self.individual_1.n_m)

        dimensions_expected = (2, 2, 0, 0)
        dimensions = self.individual_multi_1.dimensions
        self.assertEqual(dimensions, dimensions_expected)

        dimensions_expected = (2, 1, 1, 1)
        dimensions = self.individual_constr_meta.dimensions
        self.assertEqual(dimensions, dimensions_expected)

        self.assertEqual(2, self.individual_constr_meta.n_x)
        self.assertEqual(1, self.individual_constr_meta.n_f)
        self.assertEqual(1, self.individual_constr_meta.n_g)
        self.assertEqual(1, self.individual_constr_meta.n_m)

    def test_domination(self):
        self.assertFalse(self.individual_1.dominates(self.individual_2))
        self.assertTrue(self.individual_2.dominates(self.individual_1))

        self.assertFalse(self.individual_multi_1.dominates(self.individual_multi_2))
        self.assertTrue(self.individual_multi_2.dominates(self.individual_multi_1))
        self.assertFalse(self.individual_multi_3.dominates(self.individual_multi_2))

        self.assertFalse(self.individual_multi_3.dominates(self.individual_multi_2))

    def test_similarity(self):
        self.assertTrue(self.individual_1.is_similar(self.individual_2, 1e-1))
        self.assertFalse(self.individual_2.is_similar(self.individual_1, 1e-8))

        self.assertTrue(
            self.individual_multi_1.is_similar(self.individual_multi_2, 1e-1)
        )
        self.assertFalse(
            self.individual_multi_2.is_similar(self.individual_multi_1, 1e-8)
        )

    def test_to_dict(self):
        data = self.individual_1.to_dict()

        np.testing.assert_equal(data["x"], self.individual_1.x)
        np.testing.assert_equal(data["f"], self.individual_1.f)
        np.testing.assert_equal(data["x_transformed"], self.individual_1.x_transformed)
        self.assertEqual(data["variable_names"], self.individual_1.variable_names)
        self.assertEqual(
            data["independent_variable_names"],
            self.individual_1.independent_variable_names,
        )

        # Missing: Test for labels.
        # self.assertEqual(data['objective_labels'], self.individual_1.objective_labels)
        # self.assertEqual(
        #     data['nonlinear_constraint_labels'],
        #     self.individual_1.nonlinear_constraint_labels,
        # )
        # self.assertEqual(data['meta_score_labels'], elf.individual_1.meta_score_labels)

    def test_from_dict(self):
        data = self.individual_1.to_dict()
        test_individual = Individual.from_dict(data)

        np.testing.assert_equal(test_individual.x, self.individual_1.x)
        np.testing.assert_equal(test_individual.f, self.individual_1.f)
        np.testing.assert_equal(test_individual.g, self.individual_1.g)
        np.testing.assert_equal(test_individual.m, self.individual_1.m)
        np.testing.assert_equal(
            test_individual.x_transformed, self.individual_1.x_transformed
        )
        self.assertEqual(
            test_individual.variable_names, self.individual_1.variable_names
        )
        self.assertEqual(
            test_individual.independent_variable_names,
            self.individual_1.independent_variable_names,
        )
        self.assertEqual(
            test_individual.objective_labels, self.individual_1.objective_labels
        )
        self.assertEqual(
            test_individual.nonlinear_constraint_labels,
            self.individual_1.nonlinear_constraint_labels,
        )
        self.assertEqual(
            test_individual.meta_score_labels, self.individual_1.meta_score_labels
        )
        self.assertEqual(test_individual.id, self.individual_1.id)


if __name__ == "__main__":
    unittest.main()
